# pylint: disable=empty-docstring
# pylint: disable=invalid-name
# pylint: disable=line-too-long

from typing import Dict, Optional

from pylabrobot.resources.tecan.tecan_resource import TecanResource
from pylabrobot.resources.tip_rack import TipRack, TipSpot
from pylabrobot.resources.utils import create_ordered_items_2d

from .tip_creators import (
  Adapter_96_DiTi_1to1_MCA384_tip,
  Adapter_96_DiTi_4to1_MCA384_tip,
  Adapter_96_DiTi_MCA384_tip,
  Adapter_DiTi_Combo_MCA384_tip,
  Adapter_DiTi_MCA384_tip,
  Base_Nested_DiTi_MCA96_tip,
  DiTi_10ul_Filter_LiHa_L_tip,
  DiTi_10ul_Filter_LiHa_tip,
  DiTi_10ul_Filter_Nested_LiHa_tip,
  DiTi_10ul_LiHa_L_tip,
  DiTi_10ul_LiHa_tip,
  DiTi_10ul_Nested_LiHa_tip,
  DiTi_10ul_SBS_Filter_LiHa_tip,
  DiTi_10ul_SBS_LiHa_tip,
  DiTi_15ul_Filter_MCA384_tip,
  DiTi_15ul_MCA384_tip,
  DiTi_50ul_CL_Filter_LiHa_tip,
  DiTi_50ul_CL_LiHa_tip,
  DiTi_50ul_Filter_LiHa_tip,
  DiTi_50ul_Filter_MCA384_tip,
  DiTi_50ul_LiHa_tip,
  DiTi_50ul_MCA96_tip,
  DiTi_50ul_MCA384_tip,
  DiTi_50ul_Nested_MCA96_tip,
  DiTi_50ul_SBS_LiHa_tip,
  DiTi_50ul_SBS_MCA96_tip,
  DiTi_50ul_Te_MO_tip,
  DiTi_100ul_Filter_MCA96_tip,
  DiTi_100ul_Filter_Te_MO_tip,
  DiTi_100ul_MCA96_tip,
  DiTi_100ul_Nested_MCA96_tip,
  DiTi_100ul_SBS_MCA96_tip,
  DiTi_100ul_Te_MO_tip,
  DiTi_125ul_Filter_MCA384_tip,
  DiTi_125ul_MCA384_tip,
  DiTi_200ul_CL_Filter_LiHa_tip,
  DiTi_200ul_CL_LiHa_tip,
  DiTi_200ul_Filter_LiHa_tip,
  DiTi_200ul_Filter_MCA96_tip,
  DiTi_200ul_Filter_Te_MO_tip,
  DiTi_200ul_LiHa_tip,
  DiTi_200ul_MCA96_tip,
  DiTi_200ul_Nested_MCA96_tip,
  DiTi_200ul_SBS_LiHa_tip,
  DiTi_200ul_SBS_MCA96_tip,
  DiTi_200ul_Te_MO_tip,
  DiTi_200ul_w_b_filter_MCA96_tip,
  DiTi_200ul_wide_bore_MCA96_tip,
  DiTi_350ul_Nested_LiHa_tip,
  DiTi_500ul_Filter_SBS_MCA96_tip,
  DiTi_500ul_SBS_MCA96_tip,
  DiTi_1000ul_CL_Filter_LiHa_tip,
  DiTi_1000ul_CL_LiHa_tip,
  DiTi_1000ul_Filter_LiHa_tip,
  DiTi_1000ul_LiHa_tip,
  DiTi_1000ul_SBS_LiHa_tip,
  DiTi_1000ul_W_B_Filter_LiHa_tip,
  DiTi_5000ul_Filter_LiHa_tip,
  DiTi_5000ul_LiHa_tip,
  DiTi_Nested_Waste_MCA384_tip,
)


class TecanTipRack(TipRack, TecanResource):
  """Base class for Tecan tip racks."""

  def __init__(
    self,
    name: str,
    size_x: float,
    size_y: float,
    size_z: float,
    z_start: float,
    z_dispense: float,
    z_max: float,
    area: float,
    ordered_items: Optional[Dict[str, TipSpot]] = None,
    category: str = "tecan_plate",
    model: Optional[str] = None,
  ):
    super().__init__(name, size_x, size_y, size_z, ordered_items, category=category, model=model)

    self.z_start = z_start
    self.z_dispense = z_dispense
    self.z_max = z_max
    self.area = area


def DiTi_100ul_Te_MO(name: str) -> TecanTipRack:
  return TecanTipRack(
    name=name,
    size_x=128.0,
    size_y=85.8,
    size_z=15.0,
    model="DiTi_100ul_Te_MO",
    z_start=1230.0,
    z_dispense=1280.0,
    z_max=1430.0,
    area=33.2,
    ordered_items=create_ordered_items_2d(
      TipSpot,
      num_items_x=12,
      num_items_y=8,
      dx=10.0,
      dy=6.9,
      dz=42.0,
      item_dx=9.0,
      item_dy=9.0,
      size_x=9.0,
      size_y=9.0,
      make_tip=DiTi_100ul_Te_MO_tip,
    ),
  )


def DiTi_50ul_Te_MO(name: str) -> TecanTipRack:
  return TecanTipRack(
    name=name,
    size_x=128.0,
    size_y=85.8,
    size_z=14.0,
    model="DiTi_50ul_Te_MO",
    z_start=1230.0,
    z_dispense=1290.0,
    z_max=1430.0,
    area=33.2,
    ordered_items=create_ordered_items_2d(
      TipSpot,
      num_items_x=12,
      num_items_y=8,
      dx=10.0,
      dy=6.9,
      dz=28.0,
      item_dx=9.0,
      item_dy=9.0,
      size_x=9.0,
      size_y=9.0,
      make_tip=DiTi_50ul_Te_MO_tip,
    ),
  )


def DiTi_200ul_Te_MO(name: str) -> TecanTipRack:
  return TecanTipRack(
    name=name,
    size_x=128.0,
    size_y=85.8,
    size_z=14.0,
    model="DiTi_200ul_Te_MO",
    z_start=1230.0,
    z_dispense=1290.0,
    z_max=1430.0,
    area=33.2,
    ordered_items=create_ordered_items_2d(
      TipSpot,
      num_items_x=12,
      num_items_y=8,
      dx=10.0,
      dy=6.9,
      dz=42.0,
      item_dx=9.0,
      item_dy=9.0,
      size_x=9.0,
      size_y=9.0,
      make_tip=DiTi_200ul_Te_MO_tip,
    ),
  )


def DiTi_100ul_Filter_Te_MO(name: str) -> TecanTipRack:
  return TecanTipRack(
    name=name,
    size_x=128.0,
    size_y=85.8,
    size_z=6.7,
    model="DiTi_100ul_Filter_Te_MO",
    z_start=1230.0,
    z_dispense=1290.0,
    z_max=1357.0,
    area=33.2,
    ordered_items=create_ordered_items_2d(
      TipSpot,
      num_items_x=12,
      num_items_y=8,
      dx=10.0,
      dy=6.9,
      dz=42.0,
      item_dx=9.0,
      item_dy=9.0,
      size_x=9.0,
      size_y=9.0,
      make_tip=DiTi_100ul_Filter_Te_MO_tip,
    ),
  )


def DiTi_200ul_Filter_Te_MO(name: str) -> TecanTipRack:
  return TecanTipRack(
    name=name,
    size_x=128.0,
    size_y=85.8,
    size_z=14.0,
    model="DiTi_200ul_Filter_Te_MO",
    z_start=1230.0,
    z_dispense=1290.0,
    z_max=1430.0,
    area=33.2,
    ordered_items=create_ordered_items_2d(
      TipSpot,
      num_items_x=12,
      num_items_y=8,
      dx=10.0,
      dy=6.9,
      dz=42.0,
      item_dx=9.0,
      item_dy=9.0,
      size_x=9.0,
      size_y=9.0,
      make_tip=DiTi_200ul_Filter_Te_MO_tip,
    ),
  )


def Adapter_96_DiTi_MCA384(name: str) -> TecanTipRack:
  """Tecan part no. 30032043 or 30032063 - Picks 96 tips, a single row or a single column. Works with MCA384 disposable tips."""
  return TecanTipRack(
    name=name,
    size_x=123.4,
    size_y=80.4,
    size_z=3.9,
    model="Adapter_96_DiTi_MCA384",
    z_start=1422.0,
    z_dispense=1422.0,
    z_max=1461.0,
    area=33.2,
    ordered_items=create_ordered_items_2d(
      TipSpot,
      num_items_x=12,
      num_items_y=8,
      dx=7.7,
      dy=4.2,
      dz=15.9,
      item_dx=9.0,
      item_dy=9.0,
      size_x=9.0,
      size_y=9.0,
      make_tip=Adapter_96_DiTi_MCA384_tip,
    ),
  )


def Adapter_DiTi_Combo_MCA384(name: str) -> TecanTipRack:
  """Tecan part no. 30032060 - Picks 384 tips, a single row of 24 tips, a single column of 16 tips or two columns of 16 tips each."""
  return TecanTipRack(
    name=name,
    size_x=128.4,
    size_y=85.4,
    size_z=3.5,
    model="Adapter_DiTi_Combo_MCA384",
    z_start=1422.0,
    z_dispense=1422.0,
    z_max=1457.0,
    area=33.2,
    ordered_items=create_ordered_items_2d(
      TipSpot,
      num_items_x=24,
      num_items_y=16,
      dx=9.95,
      dy=6.45,
      dz=15.9,
      item_dx=4.5,
      item_dy=4.5,
      size_x=4.5,
      size_y=4.5,
      make_tip=Adapter_DiTi_Combo_MCA384_tip,
    ),
  )


def Adapter_DiTi_MCA384(name: str) -> TecanTipRack:
  """Tecan part no. 30032061 - Picks 384 tips. Works with MCA384 disposable tips (Former Part no. 30032041)."""
  return TecanTipRack(
    name=name,
    size_x=128.4,
    size_y=85.4,
    size_z=3.9,
    model="Adapter_DiTi_MCA384",
    z_start=1422.0,
    z_dispense=1422.0,
    z_max=1461.0,
    area=33.2,
    ordered_items=create_ordered_items_2d(
      TipSpot,
      num_items_x=24,
      num_items_y=16,
      dx=9.95,
      dy=6.45,
      dz=15.9,
      item_dx=4.5,
      item_dy=4.5,
      size_x=4.5,
      size_y=4.5,
      make_tip=Adapter_DiTi_MCA384_tip,
    ),
  )


def DiTi_100ul_Filter_MCA96(name: str) -> TecanTipRack:
  """Tecan part no. 10612347"""
  return TecanTipRack(
    name=name,
    size_x=120.6,
    size_y=84.2,
    size_z=20.4,
    model="DiTi_100ul_Filter_MCA96",
    z_start=1541.0,
    z_dispense=1531.0,
    z_max=1735.0,
    area=33.2,
    ordered_items=create_ordered_items_2d(
      TipSpot,
      num_items_x=12,
      num_items_y=8,
      dx=6.3,
      dy=5.8,
      dz=43.1,
      item_dx=9.0,
      item_dy=9.0,
      size_x=9.0,
      size_y=9.0,
      make_tip=DiTi_100ul_Filter_MCA96_tip,
    ),
  )


def DiTi_100ul_MCA96(name: str) -> TecanTipRack:
  """Tecan part no. 10612345 or 10612346"""
  return TecanTipRack(
    name=name,
    size_x=120.6,
    size_y=84.2,
    size_z=20.4,
    model="DiTi_100ul_MCA96",
    z_start=1541.0,
    z_dispense=1531.0,
    z_max=1735.0,
    area=33.2,
    ordered_items=create_ordered_items_2d(
      TipSpot,
      num_items_x=12,
      num_items_y=8,
      dx=6.3,
      dy=5.8,
      dz=43.1,
      item_dx=9.0,
      item_dy=9.0,
      size_x=9.0,
      size_y=9.0,
      make_tip=DiTi_100ul_MCA96_tip,
    ),
  )


def DiTi_200ul_Filter_MCA96(name: str) -> TecanTipRack:
  """Tecan part no. 10612342"""
  return TecanTipRack(
    name=name,
    size_x=120.6,
    size_y=84.2,
    size_z=20.4,
    model="DiTi_200ul_Filter_MCA96",
    z_start=1541.0,
    z_dispense=1531.0,
    z_max=1735.0,
    area=33.2,
    ordered_items=create_ordered_items_2d(
      TipSpot,
      num_items_x=12,
      num_items_y=8,
      dx=6.3,
      dy=5.8,
      dz=43.1,
      item_dx=9.0,
      item_dy=9.0,
      size_x=9.0,
      size_y=9.0,
      make_tip=DiTi_200ul_Filter_MCA96_tip,
    ),
  )


def DiTi_200ul_MCA96(name: str) -> TecanTipRack:
  """Tecan part no. 10612340 or 10612341"""
  return TecanTipRack(
    name=name,
    size_x=120.6,
    size_y=84.2,
    size_z=20.4,
    model="DiTi_200ul_MCA96",
    z_start=1541.0,
    z_dispense=1531.0,
    z_max=1735.0,
    area=33.2,
    ordered_items=create_ordered_items_2d(
      TipSpot,
      num_items_x=12,
      num_items_y=8,
      dx=6.3,
      dy=5.8,
      dz=43.1,
      item_dx=9.0,
      item_dy=9.0,
      size_x=9.0,
      size_y=9.0,
      make_tip=DiTi_200ul_MCA96_tip,
    ),
  )


def DiTi_50ul_MCA96(name: str) -> TecanTipRack:
  """Tecan part no. 10612343"""
  return TecanTipRack(
    name=name,
    size_x=120.6,
    size_y=84.2,
    size_z=20.4,
    model="DiTi_50ul_MCA96",
    z_start=1541.0,
    z_dispense=1531.0,
    z_max=1735.0,
    area=33.2,
    ordered_items=create_ordered_items_2d(
      TipSpot,
      num_items_x=12,
      num_items_y=8,
      dx=6.3,
      dy=5.8,
      dz=29.6,
      item_dx=9.0,
      item_dy=9.0,
      size_x=9.0,
      size_y=9.0,
      make_tip=DiTi_50ul_MCA96_tip,
    ),
  )


def Base_Nested_DiTi_MCA96(name: str) -> TecanTipRack:
  """Tecan part no. 30038609 or 30038614 or 30038619"""
  return TecanTipRack(
    name=name,
    size_x=124.8,
    size_y=89.2,
    size_z=-0.2,
    model="Base_Nested_DiTi_MCA96",
    z_start=3282.0,
    z_dispense=3282.0,
    z_max=3280.0,
    area=33.2,
    ordered_items=create_ordered_items_2d(
      TipSpot,
      num_items_x=4,
      num_items_y=4,
      dx=-3.6,
      dy=-3.7,
      dz=43.1,
      item_dx=33.0,
      item_dy=33.0,
      size_x=33.0,
      size_y=33.0,
      make_tip=Base_Nested_DiTi_MCA96_tip,
    ),
  )


def DiTi_100ul_Nested_MCA96(name: str) -> TecanTipRack:
  """Tecan part no. 30038614"""
  return TecanTipRack(
    name=name,
    size_x=128.2,
    size_y=85.0,
    size_z=16.6,
    model="DiTi_100ul_Nested_MCA96",
    z_start=1954.0,
    z_dispense=1933.0,
    z_max=2099.0,
    area=33.2,
    ordered_items=create_ordered_items_2d(
      TipSpot,
      num_items_x=12,
      num_items_y=8,
      dx=10.1,
      dy=6.5,
      dz=43.1,
      item_dx=9.0,
      item_dy=9.0,
      size_x=9.0,
      size_y=9.0,
      make_tip=DiTi_100ul_Nested_MCA96_tip,
    ),
  )


def DiTi_100ul_SBS_MCA96(name: str) -> TecanTipRack:
  return TecanTipRack(
    name=name,
    size_x=128.2,
    size_y=88.2,
    size_z=26.0,
    model="DiTi_100ul_SBS_MCA96",
    z_start=1538.0,
    z_dispense=1478.0,
    z_max=1738.0,
    area=33.2,
    ordered_items=create_ordered_items_2d(
      TipSpot,
      num_items_x=12,
      num_items_y=8,
      dx=10.1,
      dy=7.8,
      dz=43.1,
      item_dx=9.0,
      item_dy=9.0,
      size_x=9.0,
      size_y=9.0,
      make_tip=DiTi_100ul_SBS_MCA96_tip,
    ),
  )


def DiTi_200ul_SBS_MCA96(name: str) -> TecanTipRack:
  return TecanTipRack(
    name=name,
    size_x=128.2,
    size_y=88.2,
    size_z=26.0,
    model="DiTi_200ul_SBS_MCA96",
    z_start=1538.0,
    z_dispense=1478.0,
    z_max=1738.0,
    area=33.2,
    ordered_items=create_ordered_items_2d(
      TipSpot,
      num_items_x=12,
      num_items_y=8,
      dx=10.1,
      dy=7.8,
      dz=43.1,
      item_dx=9.0,
      item_dy=9.0,
      size_x=9.0,
      size_y=9.0,
      make_tip=DiTi_200ul_SBS_MCA96_tip,
    ),
  )


def DiTi_50ul_SBS_MCA96(name: str) -> TecanTipRack:
  return TecanTipRack(
    name=name,
    size_x=128.2,
    size_y=88.2,
    size_z=25.0,
    model="DiTi_50ul_SBS_MCA96",
    z_start=1538.0,
    z_dispense=1478.0,
    z_max=1728.0,
    area=33.2,
    ordered_items=create_ordered_items_2d(
      TipSpot,
      num_items_x=12,
      num_items_y=8,
      dx=10.1,
      dy=7.8,
      dz=29.6,
      item_dx=9.0,
      item_dy=9.0,
      size_x=9.0,
      size_y=9.0,
      make_tip=DiTi_50ul_SBS_MCA96_tip,
    ),
  )


def DiTi_50ul_Nested_MCA96(name: str) -> TecanTipRack:
  """Tecan part no. 30038609"""
  return TecanTipRack(
    name=name,
    size_x=128.2,
    size_y=85.0,
    size_z=16.6,
    model="DiTi_50ul_Nested_MCA96",
    z_start=1954.0,
    z_dispense=1933.0,
    z_max=2099.0,
    area=33.2,
    ordered_items=create_ordered_items_2d(
      TipSpot,
      num_items_x=12,
      num_items_y=8,
      dx=10.1,
      dy=6.5,
      dz=29.6,
      item_dx=9.0,
      item_dy=9.0,
      size_x=9.0,
      size_y=9.0,
      make_tip=DiTi_50ul_Nested_MCA96_tip,
    ),
  )


def Adapter_96_DiTi_1to1_MCA384(name: str) -> TecanTipRack:
  """Tecan part no. 30032048 - Picks 96 tips, single rows/columns. Uses 1channel per tip. Works with MCA96 disposable tips."""
  return TecanTipRack(
    name=name,
    size_x=128.4,
    size_y=85.4,
    size_z=3.9,
    model="Adapter_96_DiTi_1to1_MCA384",
    z_start=1422.0,
    z_dispense=1422.0,
    z_max=1461.0,
    area=33.2,
    ordered_items=create_ordered_items_2d(
      TipSpot,
      num_items_x=12,
      num_items_y=8,
      dx=7.45,
      dy=3.95,
      dz=25.2,
      item_dx=9.5,
      item_dy=9.5,
      size_x=9.5,
      size_y=9.5,
      make_tip=Adapter_96_DiTi_1to1_MCA384_tip,
    ),
  )


def DiTi_200ul_Nested_MCA96(name: str) -> TecanTipRack:
  """Tecan part no. 30038619"""
  return TecanTipRack(
    name=name,
    size_x=128.0,
    size_y=84.8,
    size_z=16.6,
    model="DiTi_200ul_Nested_MCA96",
    z_start=1954.0,
    z_dispense=1933.0,
    z_max=2099.0,
    area=33.2,
    ordered_items=create_ordered_items_2d(
      TipSpot,
      num_items_x=12,
      num_items_y=8,
      dx=10.0,
      dy=6.4,
      dz=43.1,
      item_dx=9.0,
      item_dy=9.0,
      size_x=9.0,
      size_y=9.0,
      make_tip=DiTi_200ul_Nested_MCA96_tip,
    ),
  )


def DiTi_200ul_w_b_filter_MCA96(name: str) -> TecanTipRack:
  """Tecan part no. 30050349, not for volumes under 5ul"""
  return TecanTipRack(
    name=name,
    size_x=128.2,
    size_y=88.2,
    size_z=26.0,
    model="DiTi_200ul_w_b_filter_MCA96",
    z_start=1538.0,
    z_dispense=1478.0,
    z_max=1738.0,
    area=33.2,
    ordered_items=create_ordered_items_2d(
      TipSpot,
      num_items_x=12,
      num_items_y=8,
      dx=10.1,
      dy=7.8,
      dz=43.1,
      item_dx=9.0,
      item_dy=9.0,
      size_x=9.0,
      size_y=9.0,
      make_tip=DiTi_200ul_w_b_filter_MCA96_tip,
    ),
  )


def DiTi_200ul_wide_bore_MCA96(name: str) -> TecanTipRack:
  """Tecan part no. 30050348, not for volumes under 5ul"""
  return TecanTipRack(
    name=name,
    size_x=128.2,
    size_y=88.2,
    size_z=26.0,
    model="DiTi_200ul_wide_bore_MCA96",
    z_start=1538.0,
    z_dispense=1478.0,
    z_max=1738.0,
    area=33.2,
    ordered_items=create_ordered_items_2d(
      TipSpot,
      num_items_x=12,
      num_items_y=8,
      dx=10.1,
      dy=7.8,
      dz=43.1,
      item_dx=9.0,
      item_dy=9.0,
      size_x=9.0,
      size_y=9.0,
      make_tip=DiTi_200ul_wide_bore_MCA96_tip,
    ),
  )


def Adapter_96_DiTi_4to1_MCA384(name: str) -> TecanTipRack:
  """Tecan part no. 30032042 or 30032062 - EVA (Extended Volume Adapter) - Maximum capacity 500 uL. Use with SBS 96 tip box only."""
  return TecanTipRack(
    name=name,
    size_x=129.2,
    size_y=85.4,
    size_z=3.9,
    model="Adapter_96_DiTi_4to1_MCA384",
    z_start=1422.0,
    z_dispense=1422.0,
    z_max=1461.0,
    area=33.2,
    ordered_items=create_ordered_items_2d(
      TipSpot,
      num_items_x=12,
      num_items_y=8,
      dx=7.85,
      dy=3.95,
      dz=25.2,
      item_dx=9.5,
      item_dy=9.5,
      size_x=9.5,
      size_y=9.5,
      make_tip=Adapter_96_DiTi_4to1_MCA384_tip,
    ),
  )


def DiTi_500ul_Filter_SBS_MCA96(name: str) -> TecanTipRack:
  return TecanTipRack(
    name=name,
    size_x=129.6,
    size_y=86.6,
    size_z=15.0,
    model="DiTi_500ul_Filter_SBS_MCA96",
    z_start=1453.0,
    z_dispense=1410.0,
    z_max=1560.0,
    area=33.2,
    ordered_items=create_ordered_items_2d(
      TipSpot,
      num_items_x=12,
      num_items_y=8,
      dx=10.8,
      dy=6.9,
      dz=54.0,
      item_dx=9.0,
      item_dy=9.0,
      size_x=9.0,
      size_y=9.0,
      make_tip=DiTi_500ul_Filter_SBS_MCA96_tip,
    ),
  )


def DiTi_500ul_SBS_MCA96(name: str) -> TecanTipRack:
  return TecanTipRack(
    name=name,
    size_x=129.2,
    size_y=86.8,
    size_z=14.0,
    model="DiTi_500ul_SBS_MCA96",
    z_start=1438.0,
    z_dispense=1438.0,
    z_max=1578.0,
    area=33.2,
    ordered_items=create_ordered_items_2d(
      TipSpot,
      num_items_x=12,
      num_items_y=8,
      dx=10.6,
      dy=6.9,
      dz=54.0,
      item_dx=9.0,
      item_dy=9.0,
      size_x=9.0,
      size_y=9.0,
      make_tip=DiTi_500ul_SBS_MCA96_tip,
    ),
  )


def DiTi_Nested_Waste_MCA384(name: str) -> TecanTipRack:
  """Tecan part no. 30054549"""
  return TecanTipRack(
    name=name,
    size_x=166.0,
    size_y=95.0,
    size_z=0.0,
    model="DiTi_Nested_Waste_MCA384",
    z_start=1940.0,
    z_dispense=1940.0,
    z_max=1940.0,
    area=20385.0,
    ordered_items=create_ordered_items_2d(
      TipSpot,
      num_items_x=12,
      num_items_y=8,
      dx=29.0,
      dy=11.5,
      dz=0.0,
      item_dx=9.0,
      item_dy=9.0,
      size_x=9.0,
      size_y=9.0,
      make_tip=DiTi_Nested_Waste_MCA384_tip,
    ),
  )


def DiTi_1000ul_SBS_LiHa(name: str) -> TecanTipRack:
  return TecanTipRack(
    name=name,
    size_x=128.0,
    size_y=85.8,
    size_z=25.0,
    model="DiTi_1000ul_SBS_LiHa",
    z_start=1010.0,
    z_dispense=1010.0,
    z_max=1260.0,
    area=33.2,
    ordered_items=create_ordered_items_2d(
      TipSpot,
      num_items_x=12,
      num_items_y=8,
      dx=10.0,
      dy=6.9,
      dz=32.6,
      item_dx=9.0,
      item_dy=9.0,
      size_x=9.0,
      size_y=9.0,
      make_tip=DiTi_1000ul_SBS_LiHa_tip,
    ),
  )


def DiTi_200ul_SBS_LiHa(name: str) -> TecanTipRack:
  return TecanTipRack(
    name=name,
    size_x=128.2,
    size_y=86.8,
    size_z=30.0,
    model="DiTi_200ul_SBS_LiHa",
    z_start=1360.0,
    z_dispense=1360.0,
    z_max=1660.0,
    area=33.2,
    ordered_items=create_ordered_items_2d(
      TipSpot,
      num_items_x=12,
      num_items_y=8,
      dx=10.1,
      dy=7.4,
      dz=-5.0,
      item_dx=9.0,
      item_dy=9.0,
      size_x=9.0,
      size_y=9.0,
      make_tip=DiTi_200ul_SBS_LiHa_tip,
    ),
  )


def DiTi_50ul_SBS_LiHa(name: str) -> TecanTipRack:
  return TecanTipRack(
    name=name,
    size_x=128.2,
    size_y=86.0,
    size_z=30.0,
    model="DiTi_50ul_SBS_LiHa",
    z_start=1360.0,
    z_dispense=1360.0,
    z_max=1660.0,
    area=33.2,
    ordered_items=create_ordered_items_2d(
      TipSpot,
      num_items_x=12,
      num_items_y=8,
      dx=10.1,
      dy=7.0,
      dz=-5.3,
      item_dx=9.0,
      item_dy=9.0,
      size_x=9.0,
      size_y=9.0,
      make_tip=DiTi_50ul_SBS_LiHa_tip,
    ),
  )


def DiTi_5000ul_LiHa(name: str) -> TecanTipRack:
  """Tecan part no. 30059897 (Tecan Pure). Tip is usable in a volume range of 300ul to 4850ul."""
  return TecanTipRack(
    name=name,
    size_x=129.0,
    size_y=87.4,
    size_z=30.0,
    model="DiTi_5000ul_LiHa",
    z_start=850.0,
    z_dispense=850.0,
    z_max=1150.0,
    area=50.0,
    ordered_items=create_ordered_items_2d(
      TipSpot,
      num_items_x=6,
      num_items_y=4,
      dx=10.5,
      dy=7.6,
      dz=46.6,
      item_dx=18.0,
      item_dy=18.0,
      size_x=18.0,
      size_y=18.0,
      make_tip=DiTi_5000ul_LiHa_tip,
    ),
  )


def DiTi_5000ul_Filter_LiHa(name: str) -> TecanTipRack:
  """Tecan part no. 30065423 (Tecan Pure), Tip is usable in a volume range of 300ul to 4800ul."""
  return TecanTipRack(
    name=name,
    size_x=129.0,
    size_y=87.4,
    size_z=30.0,
    model="DiTi_5000ul_Filter_LiHa",
    z_start=850.0,
    z_dispense=850.0,
    z_max=1150.0,
    area=50.0,
    ordered_items=create_ordered_items_2d(
      TipSpot,
      num_items_x=6,
      num_items_y=4,
      dx=10.5,
      dy=7.6,
      dz=46.6,
      item_dx=18.0,
      item_dy=18.0,
      size_x=18.0,
      size_y=18.0,
      make_tip=DiTi_5000ul_Filter_LiHa_tip,
    ),
  )


def DiTi_125ul_Filter_MCA384(name: str) -> TecanTipRack:
  """Tecan part no. 30051810 (Tecan Sterile). Maximum pipetting volume is 110ul."""
  return TecanTipRack(
    name=name,
    size_x=127.3,
    size_y=85.3,
    size_z=20.0,
    model="DiTi_125ul_Filter_MCA384",
    z_start=1520.0,
    z_dispense=1490.0,
    z_max=1690.0,
    area=33.2,
    ordered_items=create_ordered_items_2d(
      TipSpot,
      num_items_x=24,
      num_items_y=16,
      dx=9.65,
      dy=6.65,
      dz=47.3,
      item_dx=4.5,
      item_dy=4.5,
      size_x=4.5,
      size_y=4.5,
      make_tip=DiTi_125ul_Filter_MCA384_tip,
    ),
  )


def DiTi_125ul_MCA384(name: str) -> TecanTipRack:
  """Tecan part no. 30051808 (Tecan Pure) or 30051809 (Tecan Sterile)"""
  return TecanTipRack(
    name=name,
    size_x=127.7,
    size_y=84.9,
    size_z=20.0,
    model="DiTi_125ul_MCA384",
    z_start=1520.0,
    z_dispense=1490.0,
    z_max=1690.0,
    area=33.2,
    ordered_items=create_ordered_items_2d(
      TipSpot,
      num_items_x=24,
      num_items_y=16,
      dx=9.85,
      dy=6.45,
      dz=47.3,
      item_dx=4.5,
      item_dy=4.5,
      size_x=4.5,
      size_y=4.5,
      make_tip=DiTi_125ul_MCA384_tip,
    ),
  )


def DiTi_15ul_Filter_MCA384(name: str) -> TecanTipRack:
  """Tecan part no. 30051804 (Tecan Sterile). Maximum pipetting volume is 14.5ul."""
  return TecanTipRack(
    name=name,
    size_x=127.9,
    size_y=84.9,
    size_z=20.3,
    model="DiTi_15ul_Filter_MCA384",
    z_start=1705.0,
    z_dispense=1676.0,
    z_max=1879.0,
    area=33.2,
    ordered_items=create_ordered_items_2d(
      TipSpot,
      num_items_x=24,
      num_items_y=16,
      dx=9.95,
      dy=6.45,
      dz=28.6,
      item_dx=4.5,
      item_dy=4.5,
      size_x=4.5,
      size_y=4.5,
      make_tip=DiTi_15ul_Filter_MCA384_tip,
    ),
  )


def DiTi_15ul_MCA384(name: str) -> TecanTipRack:
  """Tecan part no. 30051802 (Tecan Pure) or 30051803 (Tecan Sterile)"""
  return TecanTipRack(
    name=name,
    size_x=127.9,
    size_y=84.9,
    size_z=20.3,
    model="DiTi_15ul_MCA384",
    z_start=1705.0,
    z_dispense=1676.0,
    z_max=1879.0,
    area=33.2,
    ordered_items=create_ordered_items_2d(
      TipSpot,
      num_items_x=24,
      num_items_y=16,
      dx=9.95,
      dy=6.45,
      dz=28.6,
      item_dx=4.5,
      item_dy=4.5,
      size_x=4.5,
      size_y=4.5,
      make_tip=DiTi_15ul_MCA384_tip,
    ),
  )


def DiTi_50ul_Filter_MCA384(name: str) -> TecanTipRack:
  """Tecan part no. 30051807 (Tecan Sterile). Maximum pipetting volume is 40ul."""
  return TecanTipRack(
    name=name,
    size_x=127.3,
    size_y=85.3,
    size_z=20.0,
    model="DiTi_50ul_Filter_MCA384",
    z_start=1520.0,
    z_dispense=1490.0,
    z_max=1690.0,
    area=33.2,
    ordered_items=create_ordered_items_2d(
      TipSpot,
      num_items_x=24,
      num_items_y=16,
      dx=9.65,
      dy=6.65,
      dz=41.4,
      item_dx=4.5,
      item_dy=4.5,
      size_x=4.5,
      size_y=4.5,
      make_tip=DiTi_50ul_Filter_MCA384_tip,
    ),
  )


def DiTi_50ul_MCA384(name: str) -> TecanTipRack:
  """Tecan part no. 30051805 (Tecan Pure) or 30051806 (Tecan Sterile)"""
  return TecanTipRack(
    name=name,
    size_x=127.3,
    size_y=85.3,
    size_z=20.0,
    model="DiTi_50ul_MCA384",
    z_start=1520.0,
    z_dispense=1490.0,
    z_max=1690.0,
    area=33.2,
    ordered_items=create_ordered_items_2d(
      TipSpot,
      num_items_x=24,
      num_items_y=16,
      dx=9.65,
      dy=6.65,
      dz=41.4,
      item_dx=4.5,
      item_dy=4.5,
      size_x=4.5,
      size_y=4.5,
      make_tip=DiTi_50ul_MCA384_tip,
    ),
  )


def DiTi_1000ul_Filter_LiHa(name: str) -> TecanTipRack:
  """Tecan part no. 10612513 or 10612555 or 30000631"""
  return TecanTipRack(
    name=name,
    size_x=123.4,
    size_y=89.4,
    size_z=22.2,
    model="DiTi_1000ul_Filter_LiHa",
    z_start=877.0,
    z_dispense=865.0,
    z_max=1087.0,
    area=33.2,
    ordered_items=create_ordered_items_2d(
      TipSpot,
      num_items_x=12,
      num_items_y=8,
      dx=7.7,
      dy=8.7,
      dz=32.6,
      item_dx=9.0,
      item_dy=9.0,
      size_x=9.0,
      size_y=9.0,
      make_tip=DiTi_1000ul_Filter_LiHa_tip,
    ),
  )


def DiTi_1000ul_LiHa(name: str) -> TecanTipRack:
  """Tecan part no. 10612554 or 30000630"""
  return TecanTipRack(
    name=name,
    size_x=123.4,
    size_y=89.4,
    size_z=22.2,
    model="DiTi_1000ul_LiHa",
    z_start=877.0,
    z_dispense=865.0,
    z_max=1087.0,
    area=33.2,
    ordered_items=create_ordered_items_2d(
      TipSpot,
      num_items_x=12,
      num_items_y=8,
      dx=7.7,
      dy=8.7,
      dz=32.6,
      item_dx=9.0,
      item_dy=9.0,
      size_x=9.0,
      size_y=9.0,
      make_tip=DiTi_1000ul_LiHa_tip,
    ),
  )


def DiTi_10ul_Filter_LiHa(name: str) -> TecanTipRack:
  """Tecan part no. 10612517"""
  return TecanTipRack(
    name=name,
    size_x=123.4,
    size_y=89.4,
    size_z=22.2,
    model="DiTi_10ul_Filter_LiHa",
    z_start=877.0,
    z_dispense=865.0,
    z_max=1087.0,
    area=33.2,
    ordered_items=create_ordered_items_2d(
      TipSpot,
      num_items_x=12,
      num_items_y=8,
      dx=7.7,
      dy=8.7,
      dz=-31.3,
      item_dx=9.0,
      item_dy=9.0,
      size_x=9.0,
      size_y=9.0,
      make_tip=DiTi_10ul_Filter_LiHa_tip,
    ),
  )


def DiTi_10ul_LiHa(name: str) -> TecanTipRack:
  """Tecan part no. 10612516"""
  return TecanTipRack(
    name=name,
    size_x=123.4,
    size_y=89.4,
    size_z=22.2,
    model="DiTi_10ul_LiHa",
    z_start=877.0,
    z_dispense=865.0,
    z_max=1087.0,
    area=33.2,
    ordered_items=create_ordered_items_2d(
      TipSpot,
      num_items_x=12,
      num_items_y=8,
      dx=7.7,
      dy=8.7,
      dz=-31.3,
      item_dx=9.0,
      item_dy=9.0,
      size_x=9.0,
      size_y=9.0,
      make_tip=DiTi_10ul_LiHa_tip,
    ),
  )


def DiTi_200ul_Filter_LiHa(name: str) -> TecanTipRack:
  """Tecan part no. 10612511 or 10612553 or 30000629"""
  return TecanTipRack(
    name=name,
    size_x=123.4,
    size_y=89.4,
    size_z=22.2,
    model="DiTi_200ul_Filter_LiHa",
    z_start=877.0,
    z_dispense=865.0,
    z_max=1087.0,
    area=33.2,
    ordered_items=create_ordered_items_2d(
      TipSpot,
      num_items_x=12,
      num_items_y=8,
      dx=7.7,
      dy=8.7,
      dz=-5.0,
      item_dx=9.0,
      item_dy=9.0,
      size_x=9.0,
      size_y=9.0,
      make_tip=DiTi_200ul_Filter_LiHa_tip,
    ),
  )


def DiTi_200ul_LiHa(name: str) -> TecanTipRack:
  """Tecan part no. 10612552 or 30000627"""
  return TecanTipRack(
    name=name,
    size_x=123.4,
    size_y=89.4,
    size_z=22.2,
    model="DiTi_200ul_LiHa",
    z_start=877.0,
    z_dispense=865.0,
    z_max=1087.0,
    area=33.2,
    ordered_items=create_ordered_items_2d(
      TipSpot,
      num_items_x=12,
      num_items_y=8,
      dx=7.7,
      dy=8.7,
      dz=-5.0,
      item_dx=9.0,
      item_dy=9.0,
      size_x=9.0,
      size_y=9.0,
      make_tip=DiTi_200ul_LiHa_tip,
    ),
  )


def DiTi_50ul_Filter_LiHa(name: str) -> TecanTipRack:
  """Tecan part no. 30032114"""
  return TecanTipRack(
    name=name,
    size_x=123.4,
    size_y=89.4,
    size_z=22.2,
    model="DiTi_50ul_Filter_LiHa",
    z_start=877.0,
    z_dispense=865.0,
    z_max=1087.0,
    area=33.2,
    ordered_items=create_ordered_items_2d(
      TipSpot,
      num_items_x=12,
      num_items_y=8,
      dx=7.7,
      dy=8.7,
      dz=-5.3,
      item_dx=9.0,
      item_dy=9.0,
      size_x=9.0,
      size_y=9.0,
      make_tip=DiTi_50ul_Filter_LiHa_tip,
    ),
  )


def DiTi_50ul_LiHa(name: str) -> TecanTipRack:
  """Tecan part no. 30032115"""
  return TecanTipRack(
    name=name,
    size_x=123.4,
    size_y=89.4,
    size_z=22.2,
    model="DiTi_50ul_LiHa",
    z_start=877.0,
    z_dispense=865.0,
    z_max=1087.0,
    area=33.2,
    ordered_items=create_ordered_items_2d(
      TipSpot,
      num_items_x=12,
      num_items_y=8,
      dx=7.7,
      dy=8.7,
      dz=-5.3,
      item_dx=9.0,
      item_dy=9.0,
      size_x=9.0,
      size_y=9.0,
      make_tip=DiTi_50ul_LiHa_tip,
    ),
  )


def DiTi_350ul_Nested_LiHa(name: str) -> TecanTipRack:
  """Tecan part no. 30083400(Tecan Pure), 30083401 (Tecan Sterile)"""
  return TecanTipRack(
    name=name,
    size_x=130.4,
    size_y=86.2,
    size_z=16.0,
    model="DiTi_350ul_Nested_LiHa",
    z_start=2015.0,
    z_dispense=2015.0,
    z_max=2175.0,
    area=33.2,
    ordered_items=create_ordered_items_2d(
      TipSpot,
      num_items_x=12,
      num_items_y=8,
      dx=11.2,
      dy=7.1,
      dz=-5.0,
      item_dx=9.0,
      item_dy=9.0,
      size_x=9.0,
      size_y=9.0,
      make_tip=DiTi_350ul_Nested_LiHa_tip,
    ),
  )


def DiTi_10ul_Filter_LiHa_L(name: str) -> TecanTipRack:
  """Tecan part no. 30104804"""
  return TecanTipRack(
    name=name,
    size_x=123.4,
    size_y=89.4,
    size_z=22.2,
    model="DiTi_10ul_Filter_LiHa_L",
    z_start=877.0,
    z_dispense=865.0,
    z_max=1087.0,
    area=33.2,
    ordered_items=create_ordered_items_2d(
      TipSpot,
      num_items_x=12,
      num_items_y=8,
      dx=7.7,
      dy=8.7,
      dz=-28.1,
      item_dx=9.0,
      item_dy=9.0,
      size_x=9.0,
      size_y=9.0,
      make_tip=DiTi_10ul_Filter_LiHa_L_tip,
    ),
  )


def DiTi_10ul_Filter_Nested_LiHa(name: str) -> TecanTipRack:
  """Tecan part no. 30104978 (Tecan Pure), 30104979 (Tecan Sterile)"""
  return TecanTipRack(
    name=name,
    size_x=127.6,
    size_y=86.0,
    size_z=16.0,
    model="DiTi_10ul_Filter_Nested_LiHa",
    z_start=2015.0,
    z_dispense=2015.0,
    z_max=2175.0,
    area=33.2,
    ordered_items=create_ordered_items_2d(
      TipSpot,
      num_items_x=12,
      num_items_y=8,
      dx=9.8,
      dy=7.0,
      dz=-28.1,
      item_dx=9.0,
      item_dy=9.0,
      size_x=9.0,
      size_y=9.0,
      make_tip=DiTi_10ul_Filter_Nested_LiHa_tip,
    ),
  )


def DiTi_10ul_LiHa_L(name: str) -> TecanTipRack:
  """Tecan part no. 30104803"""
  return TecanTipRack(
    name=name,
    size_x=123.4,
    size_y=89.4,
    size_z=22.2,
    model="DiTi_10ul_LiHa_L",
    z_start=877.0,
    z_dispense=865.0,
    z_max=1087.0,
    area=33.2,
    ordered_items=create_ordered_items_2d(
      TipSpot,
      num_items_x=12,
      num_items_y=8,
      dx=7.7,
      dy=8.7,
      dz=-28.1,
      item_dx=9.0,
      item_dy=9.0,
      size_x=9.0,
      size_y=9.0,
      make_tip=DiTi_10ul_LiHa_L_tip,
    ),
  )


def DiTi_10ul_Nested_LiHa(name: str) -> TecanTipRack:
  """Tecan part no. 30104977 (Tecan Pure)"""
  return TecanTipRack(
    name=name,
    size_x=127.6,
    size_y=87.4,
    size_z=16.0,
    model="DiTi_10ul_Nested_LiHa",
    z_start=2015.0,
    z_dispense=2015.0,
    z_max=2175.0,
    area=33.2,
    ordered_items=create_ordered_items_2d(
      TipSpot,
      num_items_x=12,
      num_items_y=8,
      dx=9.8,
      dy=7.7,
      dz=-28.1,
      item_dx=9.0,
      item_dy=9.0,
      size_x=9.0,
      size_y=9.0,
      make_tip=DiTi_10ul_Nested_LiHa_tip,
    ),
  )


def DiTi_10ul_SBS_Filter_LiHa(name: str) -> TecanTipRack:
  """Tecan part no. 30104976 (Tecan Sterile), 30104974 (Tecan Pure, Refill)"""
  return TecanTipRack(
    name=name,
    size_x=128.2,
    size_y=86.8,
    size_z=30.0,
    model="DiTi_10ul_SBS_Filter_LiHa",
    z_start=1360.0,
    z_dispense=1360.0,
    z_max=1660.0,
    area=33.2,
    ordered_items=create_ordered_items_2d(
      TipSpot,
      num_items_x=12,
      num_items_y=8,
      dx=10.1,
      dy=7.4,
      dz=-28.1,
      item_dx=9.0,
      item_dy=9.0,
      size_x=9.0,
      size_y=9.0,
      make_tip=DiTi_10ul_SBS_Filter_LiHa_tip,
    ),
  )


def DiTi_10ul_SBS_LiHa(name: str) -> TecanTipRack:
  """Tecan part no. 30104975 (Tecan Sterile), 30104973 (Tecan Pure, Refill)"""
  return TecanTipRack(
    name=name,
    size_x=129.0,
    size_y=83.6,
    size_z=30.0,
    model="DiTi_10ul_SBS_LiHa",
    z_start=1360.0,
    z_dispense=1360.0,
    z_max=1660.0,
    area=33.2,
    ordered_items=create_ordered_items_2d(
      TipSpot,
      num_items_x=12,
      num_items_y=8,
      dx=10.5,
      dy=5.8,
      dz=-28.1,
      item_dx=9.0,
      item_dy=9.0,
      size_x=9.0,
      size_y=9.0,
      make_tip=DiTi_10ul_SBS_LiHa_tip,
    ),
  )


def DiTi_1000ul_W_B_Filter_LiHa(name: str) -> TecanTipRack:
  """Tecan part no. 30115239"""
  return TecanTipRack(
    name=name,
    size_x=123.4,
    size_y=91.2,
    size_z=22.2,
    model="DiTi_1000ul_W_B_Filter_LiHa",
    z_start=877.0,
    z_dispense=865.0,
    z_max=1087.0,
    area=33.2,
    ordered_items=create_ordered_items_2d(
      TipSpot,
      num_items_x=12,
      num_items_y=8,
      dx=7.7,
      dy=9.6,
      dz=32.0,
      item_dx=9.0,
      item_dy=9.0,
      size_x=9.0,
      size_y=9.0,
      make_tip=DiTi_1000ul_W_B_Filter_LiHa_tip,
    ),
  )


def DiTi_1000ul_CL_Filter_LiHa(name: str) -> TecanTipRack:
  """Tecan part no. 30126020 or 30126095"""
  return TecanTipRack(
    name=name,
    size_x=127.6,
    size_y=88.2,
    size_z=22.2,
    model="DiTi_1000ul_CL_Filter_LiHa",
    z_start=877.0,
    z_dispense=865.0,
    z_max=1087.0,
    area=33.2,
    ordered_items=create_ordered_items_2d(
      TipSpot,
      num_items_x=12,
      num_items_y=8,
      dx=9.8,
      dy=8.1,
      dz=32.6,
      item_dx=9.0,
      item_dy=9.0,
      size_x=9.0,
      size_y=9.0,
      make_tip=DiTi_1000ul_CL_Filter_LiHa_tip,
    ),
  )


def DiTi_1000ul_CL_LiHa(name: str) -> TecanTipRack:
  """Tecan part no. 30126019 or 30126094"""
  return TecanTipRack(
    name=name,
    size_x=127.6,
    size_y=88.2,
    size_z=22.2,
    model="DiTi_1000ul_CL_LiHa",
    z_start=877.0,
    z_dispense=865.0,
    z_max=1087.0,
    area=33.2,
    ordered_items=create_ordered_items_2d(
      TipSpot,
      num_items_x=12,
      num_items_y=8,
      dx=9.8,
      dy=8.1,
      dz=32.6,
      item_dx=9.0,
      item_dy=9.0,
      size_x=9.0,
      size_y=9.0,
      make_tip=DiTi_1000ul_CL_LiHa_tip,
    ),
  )


def DiTi_200ul_CL_Filter_LiHa(name: str) -> TecanTipRack:
  """Tecan part no. 30126018 or 30126093"""
  return TecanTipRack(
    name=name,
    size_x=129.4,
    size_y=87.4,
    size_z=22.2,
    model="DiTi_200ul_CL_Filter_LiHa",
    z_start=877.0,
    z_dispense=865.0,
    z_max=1087.0,
    area=33.2,
    ordered_items=create_ordered_items_2d(
      TipSpot,
      num_items_x=12,
      num_items_y=8,
      dx=10.7,
      dy=7.7,
      dz=-5.0,
      item_dx=9.0,
      item_dy=9.0,
      size_x=9.0,
      size_y=9.0,
      make_tip=DiTi_200ul_CL_Filter_LiHa_tip,
    ),
  )


def DiTi_200ul_CL_LiHa(name: str) -> TecanTipRack:
  """Tecan part no. 30126017 or 30126092"""
  return TecanTipRack(
    name=name,
    size_x=129.4,
    size_y=87.4,
    size_z=22.2,
    model="DiTi_200ul_CL_LiHa",
    z_start=877.0,
    z_dispense=865.0,
    z_max=1087.0,
    area=33.2,
    ordered_items=create_ordered_items_2d(
      TipSpot,
      num_items_x=12,
      num_items_y=8,
      dx=10.7,
      dy=7.7,
      dz=-5.0,
      item_dx=9.0,
      item_dy=9.0,
      size_x=9.0,
      size_y=9.0,
      make_tip=DiTi_200ul_CL_LiHa_tip,
    ),
  )


def DiTi_50ul_CL_Filter_LiHa(name: str) -> TecanTipRack:
  """Tecan part no. 30126097"""
  return TecanTipRack(
    name=name,
    size_x=123.4,
    size_y=89.4,
    size_z=22.2,
    model="DiTi_50ul_CL_Filter_LiHa",
    z_start=877.0,
    z_dispense=865.0,
    z_max=1087.0,
    area=33.2,
    ordered_items=create_ordered_items_2d(
      TipSpot,
      num_items_x=12,
      num_items_y=8,
      dx=7.7,
      dy=8.7,
      dz=-5.3,
      item_dx=9.0,
      item_dy=9.0,
      size_x=9.0,
      size_y=9.0,
      make_tip=DiTi_50ul_CL_Filter_LiHa_tip,
    ),
  )


def DiTi_50ul_CL_LiHa(name: str) -> TecanTipRack:
  """Tecan part no. 30126096"""
  return TecanTipRack(
    name=name,
    size_x=123.4,
    size_y=89.4,
    size_z=22.2,
    model="DiTi_50ul_CL_LiHa",
    z_start=877.0,
    z_dispense=865.0,
    z_max=1087.0,
    area=33.2,
    ordered_items=create_ordered_items_2d(
      TipSpot,
      num_items_x=12,
      num_items_y=8,
      dx=7.7,
      dy=8.7,
      dz=-5.3,
      item_dx=9.0,
      item_dy=9.0,
      size_x=9.0,
      size_y=9.0,
      make_tip=DiTi_50ul_CL_LiHa_tip,
    ),
  )



def matrix_multiply_3x3(A, B):
  """Multiplies two 3x3 matrices A and B."""
  return [[sum(A[i][k] * B[k][j] for k in range(3)) for j in range(3)] for i in range(3)]


def matrix_vector_multiply_3x3(A, v):
  """Multiplies a 3x3 matrix A with a 3x1 vector v."""
  return [sum(A[i][j] * v[j] for j in range(3)) for i in range(3)]



from __future__ import annotations

import functools
import sys
from abc import ABC
from typing import Any, Awaitable, Callable, TypeVar

from pylabrobot.machines.backend import MachineBackend
from pylabrobot.serializer import SerializableMixin

if sys.version_info < (3, 10):
  from typing_extensions import ParamSpec
else:
  from typing import ParamSpec

_P = ParamSpec("_P")
_R = TypeVar("_R", bound=Awaitable[Any])


def need_setup_finished(func: Callable[_P, _R]) -> Callable[_P, _R]:
  """Decorator for methods that require the machine to be set up.

  Checked by verifying `self.setup_finished` is `True`.

  Raises:
    RuntimeError: If the machine is not set up.
  """

  @functools.wraps(func)
  async def wrapper(*args, **kwargs):
    assert isinstance(args[0], Machine), "The first argument must be a Machine."
    self = args[0]

    if not self.setup_finished:
      raise RuntimeError("The setup has not finished. See `setup`.")
    return await func(*args, **kwargs)

  return wrapper


class Machine(SerializableMixin, ABC):
  """Abstract base class for machine frontends."""

  def __init__(self, backend: MachineBackend):
    self.backend = backend
    self._setup_finished = False

  @property
  def setup_finished(self) -> bool:
    return self._setup_finished

  def serialize(self) -> dict:
    return {"backend": self.backend.serialize()}

  @classmethod
  def deserialize(cls, data: dict):
    data_copy = data.copy()  # copy data because we will be modifying it
    backend_data = data_copy.pop("backend")
    backend = MachineBackend.deserialize(backend_data)
    data_copy["backend"] = backend
    return cls(**data_copy)

  async def setup(self, **backend_kwargs):
    await self.backend.setup(**backend_kwargs)
    self._setup_finished = True

  @need_setup_finished
  async def stop(self):
    await self.backend.stop()
    self._setup_finished = False

  async def __aenter__(self):
    await self.setup()
    return self

  async def __aexit__(self, exc_type, exc_value, traceback):
    await self.stop()



from pylabrobot.resources.carrier import (
  Coordinate,
  ResourceHolder,
  TroughCarrier,
  create_homogeneous_resources,
)


def Trough_CAR_4R200_A00(name: str) -> TroughCarrier:
  """Hamilton cat. no.: 185436
  Hamilton name: 'RGT_CAR_4R200_A00'.
  Trough carrier for 4x 200ml troughs. 2 tracks(T) wide.
  - Material: ? (recognisable via cLLD)
  - Sterilization_compatibility: ?

  carrier_site_pedestal_top = 134
  carrier_site_pedestal_bottom = 132.5
  pedestal_z_height = 1.5
  true_dz = 1.2
  trough_z_thickness = 1.4
  """
  return TroughCarrier(
    name=name,
    size_x=45.0,
    size_y=497.0,
    size_z=71.5,
    # pedestal_size_z=1.5 mm
    sites=create_homogeneous_resources(
      klass=ResourceHolder,
      locations=[
        Coordinate(4.0, 2.0, 34.0 + 1.5),
        Coordinate(4.0, 123.0, 34.0 + 1.5),  # TODO: properly define troughs to remove dependency
        Coordinate(4.0, 245.0, 34.0 + 1.5),  # on this 1.5mm offset (material_z_thickness)
        Coordinate(4.0, 366.0, 34.0 + 1.5),
      ],
      resource_size_x=37.0,
      resource_size_y=118.0,
      name_prefix=name,
    ),
    model="Trough_CAR_4R200_A00",
  )


def Trough_CAR_5R60_A00(name: str) -> TroughCarrier:
  """Hamilton cat. no.: 53646-01
  Hamilton name: 'RGT_CAR5X60'.
  Trough carrier for 5x 60ml troughs. 1 tracks(T) wide.
  Carries hamilton_1_trough_60mL_Vb
  """
  return TroughCarrier(
    name=name,
    size_x=22.5,  # standard
    size_y=497.0,  # standard
    size_z=104,  # measured
    sites=create_homogeneous_resources(
      klass=ResourceHolder,
      locations=[
        Coordinate(1.5, 7.0, 62.0 + 1.5),
        Coordinate(1.5, 103.0, 62.0 + 1.5),
        Coordinate(1.5, 199.0, 62.0 + 1.5),
        Coordinate(1.5, 302.0, 62.0 + 1.5),
      ],  # measured 62 to bottom of holder, but there is a 1.5mm pedestal
      resource_size_x=19.0,
      resource_size_y=90.0,
      # pedestal_size_z=1.5, # measured
      name_prefix=name,
    ),
    model=Trough_CAR_5R60_A00.__name__,
  )



"""Hamilton backends for liquid handling."""

from .base import HamiltonLiquidHandler
from .pump import Pump  # TODO: move elsewhere.
from .STAR_backend import STAR
from .vantage_backend import Vantage



def hex_to_binary(hex_str: str) -> str:
  """
  >>> hex_to_binary('01')
  '00000001'
  """
  return bin(int(hex_str, base=16))[2:].zfill(8)


def hex_to_base_twelve(hex_str: str) -> str:
  return bin(int(hex_str, base=12))[2:].zfill(15)


def validate_storage_location_number(storage_location_number: str):
  try:
    int(storage_location_number)
  except ValueError as exc:
    raise ValueError("Storage location number must be an integer.") from exc
  if len(storage_location_number) != 3:
    raise ValueError("Storage location number must be a three-digit number.")



from pylabrobot.heating_shaking import HeaterShakerBackend
from pylabrobot.shaking import ShakerChatterboxBackend
from pylabrobot.temperature_controlling import TemperatureControllerChatterboxBackend


class HeaterShakerChatterboxBackend(
  HeaterShakerBackend, ShakerChatterboxBackend, TemperatureControllerChatterboxBackend
):
  pass



import unittest
from unittest.mock import AsyncMock, MagicMock

from pylabrobot.arms.backend import SCARABackend
from pylabrobot.arms.precise_flex.coords import PreciseFlexCartesianCoords
from pylabrobot.arms.scara import ExperimentalSCARA
from pylabrobot.resources import Coordinate, Rotation


class TestExperimentalSCARA(unittest.IsolatedAsyncioTestCase):
  async def asyncSetUp(self):
    self.mock_backend = MagicMock(spec=SCARABackend)
    for method_name in [
      "move_to",
      "get_joint_position",
      "get_cartesian_position",
      "open_gripper",
      "close_gripper",
      "is_gripper_closed",
      "halt",
      "home",
      "move_to_safe",
      "approach",
      "pick_up_resource",
      "drop_resource",
    ]:
      setattr(self.mock_backend, method_name, AsyncMock())
    self.scara = ExperimentalSCARA(backend=self.mock_backend)

  async def test_move_to(self):
    position = PreciseFlexCartesianCoords(
      location=Coordinate(x=100, y=200, z=300), rotation=Rotation(x=0, y=0, z=0)
    )
    await self.scara.move_to(position)
    self.mock_backend.move_to.assert_called_once_with(position)

  async def test_get_joint_position(self):
    await self.scara.get_joint_position()
    self.mock_backend.get_joint_position.assert_called_once()

  async def test_get_cartesian_position(self):
    await self.scara.get_cartesian_position()
    self.mock_backend.get_cartesian_position.assert_called_once()

  async def test_open_gripper(self):
    gripper_width = 50.0
    await self.scara.open_gripper(gripper_width=gripper_width)
    self.mock_backend.open_gripper.assert_called_once_with(gripper_width=gripper_width)

  async def test_close_gripper(self):
    gripper_width = 50.0
    await self.scara.close_gripper(gripper_width=gripper_width)
    self.mock_backend.close_gripper.assert_called_once_with(gripper_width=gripper_width)

  async def test_is_gripper_closed(self):
    await self.scara.is_gripper_closed()
    self.mock_backend.is_gripper_closed.assert_called_once()

  async def test_halt(self):
    await self.scara.halt()
    self.mock_backend.halt.assert_called_once()

  async def test_home(self):
    await self.scara.home()
    self.mock_backend.home.assert_called_once()

  async def test_move_to_safe(self):
    await self.scara.move_to_safe()
    self.mock_backend.move_to_safe.assert_called_once()

  async def test_approach(self):
    position = PreciseFlexCartesianCoords(
      location=Coordinate(x=100, y=200, z=300), rotation=Rotation(x=0, y=0, z=0)
    )
    await self.scara.approach(position)
    self.mock_backend.approach.assert_called_once_with(position, access=None)

  async def test_pick_up_resource(self):
    position = PreciseFlexCartesianCoords(
      location=Coordinate(x=100, y=200, z=300), rotation=Rotation(x=0, y=0, z=0)
    )
    plate_width = 10.0
    await self.scara.pick_up_resource(position, plate_width)
    self.mock_backend.pick_up_resource.assert_called_once_with(
      plate_width=plate_width, position=position, access=None
    )

  async def test_drop_resource(self):
    position = PreciseFlexCartesianCoords(
      location=Coordinate(x=100, y=200, z=300), rotation=Rotation(x=0, y=0, z=0)
    )
    await self.scara.drop_resource(position)
    self.mock_backend.drop_resource.assert_called_once_with(position, access=None)



import logging
import math
import time
from typing import Any, Awaitable, Callable, Coroutine, Dict, Literal, Optional, Tuple, Union, cast

from pylabrobot.machines import Machine, need_setup_finished
from pylabrobot.plate_reading.backend import ImagerBackend
from pylabrobot.plate_reading.standard import (
  AutoExposure,
  AutoFocus,
  Exposure,
  FocalPosition,
  Gain,
  Image,
  ImagingMode,
  ImagingResult,
  NoPlateError,
  Objective,
)
from pylabrobot.resources import Plate, Resource, Rotation, Well

try:
  import cv2  # type: ignore

  CV2_AVAILABLE = True
except ImportError as e:
  cv2 = None  # type: ignore
  CV2_AVAILABLE = False
  _CV2_IMPORT_ERROR = e

try:
  import numpy as np  # type: ignore
except ImportError:
  np = None  # type: ignore[assignment]


logger = logging.getLogger(__name__)


async def _golden_ratio_search(
  func: Callable[..., Coroutine[Any, Any, float]], a: float, b: float, tol: float, timeout: float
):
  """Golden ratio search to maximize a unimodal function `func` over the interval [a, b]."""
  # thanks chat
  phi = (1 + np.sqrt(5)) / 2  # Golden ratio

  c = b - (b - a) / phi
  d = a + (b - a) / phi

  cache: Dict[float, float] = {}

  async def cached_func(x: float) -> float:
    x = round(x / tol) * tol  # round x to units of tol
    if x not in cache:
      cache[x] = await func(x)
    return cache[x]

  t0 = time.time()
  iteration = 0
  while abs(b - a) > tol:
    if (await cached_func(c)) > (await cached_func(d)):
      b = d
    else:
      a = c
    c = b - (b - a) / phi
    d = a + (b - a) / phi
    if time.time() - t0 > timeout:
      raise TimeoutError("Timeout while searching for optimal focus position")
    iteration += 1
    logger.debug("Golden ratio search (autofocus) iteration %d, a=%s, b=%s", iteration, a, b)

  return (b + a) / 2


class Imager(Resource, Machine):
  """Microscope"""

  def __init__(
    self,
    name: str,
    size_x: float,
    size_y: float,
    size_z: float,
    backend: ImagerBackend,
    rotation: Optional[Rotation] = None,
    category: Optional[str] = None,
    model: Optional[str] = None,
  ):
    Resource.__init__(
      self,
      name=name,
      size_x=size_x,
      size_y=size_y,
      size_z=size_z,
      rotation=rotation,
      category=category,
      model=model,
    )
    Machine.__init__(self, backend=backend)
    self.backend: ImagerBackend = backend  # fix type

    self.register_will_assign_resource_callback(self._will_assign_resource)

  def _will_assign_resource(self, resource: Resource):
    if len(self.children) >= 1:
      raise ValueError(
        f"Imager {self} already has a plate assigned (attempting to assign {resource})"
      )

  def get_plate(self) -> Plate:
    if len(self.children) == 0:
      raise NoPlateError("There is no plate in the plate reader.")
    return cast(Plate, self.children[0])

  async def _capture_auto_exposure(
    self,
    well: Union[Well, Tuple[int, int]],
    mode: ImagingMode,
    objective: Objective,
    auto_exposure: AutoExposure,
    focal_height: float,
    gain: float,
    **backend_kwargs,
  ) -> ImagingResult:
    """
    Capture an image with auto exposure.

    This function will iteratively adjust the exposure time until a good exposure is found.
    It uses the provided `evaluate_exposure` function to determine if the exposure is good, too high, or too low.
    It uses a weighted binary search to find the optimal exposure time. The search is weighted by exposure time,
    meaning that instead of splitting the range in half, we split the range at the point that equalizes the integral
    of the exposure time on both sides (this works out to be equal to the root mean square of the endpoints).
    """

    if focal_height == "auto":
      raise ValueError("Focal height must be specified for auto exposure")
    if gain == "auto":
      raise ValueError("Gain must be specified for auto exposure")

    def _rms_split(low: float, high: float) -> float:
      """Split point that equalizes ∫t dt on both sides (RMS of endpoints)."""
      if low == high:
        return low
      return math.sqrt((low**2 + high**2) / 2)

    low, high = auto_exposure.low, auto_exposure.high

    rounds = 0
    while high - low > 1e-3:
      if auto_exposure.max_rounds is not None and rounds >= auto_exposure.max_rounds:
        raise ValueError("Exceeded maximum number of rounds")
      rounds += 1

      p = _rms_split(low, high)
      res = await self.capture(
        well=well,
        mode=mode,
        objective=objective,
        exposure_time=p,
        focal_height=focal_height,
        gain=gain,
        **backend_kwargs,
      )
      assert len(res.images) == 1, "Expected exactly one image to be returned"
      im = res.images[0]
      evaluation = await auto_exposure.evaluate_exposure(im)

      if evaluation == "good":
        return res
      if evaluation == "lower":
        high = p
      elif evaluation == "higher":
        low = p
      else:
        raise ValueError(f"Unexpected evaluation result: {evaluation}")

    raise RuntimeError("Failed to find a good exposure time.")

  async def _capture_auto_focus(
    self,
    well: Union[Well, Tuple[int, int]],
    mode: ImagingMode,
    objective: Objective,
    exposure_time: float,
    auto_focus: AutoFocus,
    gain: float,
    **backend_kwargs,
  ) -> ImagingResult:
    async def local_capture(focal_height: float) -> ImagingResult:
      return await self.capture(
        well=well,
        mode=mode,
        objective=objective,
        exposure_time=exposure_time,
        focal_height=focal_height,
        gain=gain,
        **backend_kwargs,
      )

    async def capture_and_evaluate(focal_height: float) -> float:
      res = await local_capture(focal_height)
      return auto_focus.evaluate_focus(res.images[0])

    # Use golden ratio search to find the best focus value
    best_focal_height = await _golden_ratio_search(
      func=capture_and_evaluate,
      a=auto_focus.low,
      b=auto_focus.high,
      tol=auto_focus.tolerance,  # 1 micron
      timeout=auto_focus.timeout,
    )
    return await local_capture(best_focal_height)

  @need_setup_finished
  async def capture(
    self,
    well: Union[Well, Tuple[int, int]],
    mode: ImagingMode,
    objective: Objective,
    exposure_time: Union[Exposure, AutoExposure] = "machine-auto",
    focal_height: FocalPosition = "machine-auto",
    gain: Gain = "machine-auto",
    **backend_kwargs,
  ) -> ImagingResult:
    if exposure_time != "machine-auto" and not isinstance(
      exposure_time, (int, float, AutoExposure)
    ):
      raise TypeError(f"Invalid exposure time: {exposure_time}")
    if (
      not isinstance(focal_height, (int, float))
      and focal_height != "machine-auto"
      and not isinstance(focal_height, AutoFocus)
    ):
      raise TypeError(f"Invalid focal height: {focal_height}")

    if isinstance(well, tuple):
      row, column = well
    else:
      idx = cast(Plate, well.parent).index_of_item(well)
      if idx is None:
        raise ValueError(f"Well {well} not in plate {well.parent}")
      row, column = divmod(idx, cast(Plate, well.parent).num_items_x)

    if isinstance(exposure_time, AutoExposure):
      assert focal_height != "machine-auto", "Focal height must be specified for auto exposure"
      assert gain != "machine-auto", "Gain must be specified for auto exposure"
      return await self._capture_auto_exposure(
        well=well,
        mode=mode,
        objective=objective,
        auto_exposure=exposure_time,
        focal_height=focal_height,
        gain=gain,
        **backend_kwargs,
      )

    if isinstance(focal_height, AutoFocus):
      assert isinstance(exposure_time, (int, float)), (
        "Exposure time must be specified for auto focus"
      )
      assert gain != "machine-auto", "Gain must be specified for auto focus"
      return await self._capture_auto_focus(
        well=well,
        mode=mode,
        objective=objective,
        exposure_time=exposure_time,
        auto_focus=focal_height,
        gain=gain,
        **backend_kwargs,
      )

    return await self.backend.capture(
      row=row,
      column=column,
      mode=mode,
      objective=objective,
      exposure_time=exposure_time,
      focal_height=focal_height,
      gain=gain,
      plate=self.get_plate(),
      **backend_kwargs,
    )


def max_pixel_at_fraction(
  fraction: float, margin: float
) -> Callable[[Image], Awaitable[Literal["higher", "lower", "good"]]]:
  """The maximum pixel value in a given image should be a fraction of the maximum possible pixel value (eg 255 for 8-bit images).

  Args:
    fraction: the desired fraction of the actual maximum pixel value over the theoretically maximum pixel value (e.g. 0.8 for 80%). If it is an 8-bit image, the maximum value would be 0.8 * 255 = 204.
    margin: the margin of error that is accepted. A fraction of the theoretical maximum pixel value, e.g. 0.05 for 5%, so the maximum pixel value should be between 0.75 * 255 and 0.85 * 255.
  """

  if np is None:
    raise ImportError("numpy is required for max_pixel_at_fraction")

  async def evaluate_exposure(im) -> Literal["higher", "lower", "good"]:
    array = np.array(im, dtype=np.float32)
    value = np.max(array) - (255.0 * fraction)
    margin_value = 255.0 * margin
    if abs(value) <= margin_value:
      return "good"
    # lower the exposure time if the max pixel value is too high
    return "lower" if value > 0 else "higher"

  return evaluate_exposure


def fraction_overexposed(
  fraction: float, margin: float, max_pixel_value: int = 255
) -> Callable[[Image], Awaitable[Literal["higher", "lower", "good"]]]:
  """A certain fraction of pixels in the image should be overexposed (e.g. 0.5%).

  This is useful for images that are not well illuminated, as it ensures that a certain fraction of pixels is overexposed, which can help with image quality.

  Args:
    fraction: the desired fraction of pixels that should be overexposed (e.g. 0.005 for 0.5%). Overexposed is defined as pixels with a value greater than the maximum pixel value (e.g. 255 for 8-bit images). You can customize this number if needed.
    margin: the margin of error for the fraction of pixels that should be overexposed (e.g. 0.001 for 0.1%, so the fraction of overexposed pixels should be between 0.004 and 0.006).
    max_pixel_value: the maximum pixel value for the image (e.g. 255 for 8-bit images). You can override it to change the definition of "overexposed" pixels.
  """

  if np is None:
    raise ImportError("numpy is required for fraction_overexposed")

  async def evaluate_exposure(im) -> Literal["higher", "lower", "good"]:
    # count the number of pixels that are overexposed
    arr = np.asarray(im, dtype=np.uint8)
    actual_fraction = np.count_nonzero(arr > max_pixel_value) / arr.size
    lower_bound, upper_bound = fraction - margin, fraction + margin
    if lower_bound <= actual_fraction <= upper_bound:
      return "good"
    # too many saturated pixels -> shorten exposure
    return "lower" if (actual_fraction - fraction) > 0 else "higher"

  return evaluate_exposure


def evaluate_focus_nvmg_sobel(image: Image) -> float:
  """Evaluate the focus of an image using the Normalized Variance of the Gradient Magnitude (NVMG) method with Sobel filters.

  I think Chat invented this method.

  Only uses the center 50% of the image to avoid edge effects.
  """
  if not CV2_AVAILABLE:
    raise RuntimeError(
      f"cv2 needs to be installed for auto focus. Import error: {_CV2_IMPORT_ERROR}"
    )

  # cut out 25% on each side
  np_image = np.array(image, dtype=np.float64)
  height, width = np_image.shape[:2]
  crop_height = height // 4
  crop_width = width // 4
  np_image = np_image[crop_height : height - crop_height, crop_width : width - crop_width]

  # NVMG: Normalized Variance of the Gradient Magnitude
  # Chat invented this i think
  sobel_x = cv2.Sobel(np_image, cv2.CV_64F, 1, 0, ksize=3)
  sobel_y = cv2.Sobel(np_image, cv2.CV_64F, 0, 1, ksize=3)
  gradient_magnitude = np.sqrt(sobel_x**2 + sobel_y**2)

  mean_gm = np.mean(gradient_magnitude)
  var_gm = np.var(gradient_magnitude)
  sharpness = var_gm / (mean_gm + 1e-6)
  return cast(float, sharpness)



"""ML Star tip carriers"""

from pylabrobot.resources.carrier import (
  ResourceHolder,
  TipCarrier,
  create_homogeneous_resources,
)
from pylabrobot.resources.coordinate import Coordinate


def TIP_CAR_120BC_4mlTF_A00(name: str) -> TipCarrier:
  """Tip carrier with 5 4ml tip with filter racks landscape"""
  return TipCarrier(
    name=name,
    size_x=135.0,
    size_y=497.0,
    size_z=130.0,
    sites=create_homogeneous_resources(
      klass=ResourceHolder,
      locations=[
        Coordinate(6.2, 10.0, 114.95),
        Coordinate(6.2, 106.0, 114.95),
        Coordinate(6.2, 202.0, 114.95),
        Coordinate(6.2, 298.0, 114.95),
        Coordinate(6.2, 394.0, 114.95),
      ],
      resource_size_x=122.4,
      resource_size_y=82.6,
      name_prefix=name,
    ),
    model="TIP_CAR_120BC_4mlTF_A00",
  )


def TIP_CAR_120BC_5mlT_A00(name: str) -> TipCarrier:
  """Tip carrier with 5 5ml tip racks landscape"""
  return TipCarrier(
    name=name,
    size_x=135.0,
    size_y=497.0,
    size_z=130.0,
    sites=create_homogeneous_resources(
      klass=ResourceHolder,
      locations=[
        Coordinate(6.2, 10.0, 114.95),
        Coordinate(6.2, 106.0, 114.95),
        Coordinate(6.2, 202.0, 114.95),
        Coordinate(6.2, 298.0, 114.95),
        Coordinate(6.2, 394.0, 114.95),
      ],
      resource_size_x=122.4,
      resource_size_y=82.6,
      name_prefix=name,
    ),
    model="TIP_CAR_120BC_5mlT_A00",
  )


def TIP_CAR_288_A00(name: str) -> TipCarrier:
  """Carrier for 3 Racks with 96 Tips portrait  [revision A00]"""
  return TipCarrier(
    name=name,
    size_x=135.0,
    size_y=497.0,
    size_z=130.0,
    sites=create_homogeneous_resources(
      klass=ResourceHolder,
      locations=[
        Coordinate(26.3, 36.3, 114.9),
        Coordinate(26.3, 182.213, 114.9),
        Coordinate(26.3, 328.213, 114.9),
      ],
      resource_size_x=82.6,
      resource_size_y=122.4,
      name_prefix=name,
    ),
    model="TIP_CAR_288_A00",
  )


def TIP_CAR_288_B00(name: str) -> TipCarrier:
  """Carrier for 3 Racks with 96 Tips portrait [revision B00]"""
  return TipCarrier(
    name=name,
    size_x=112.5,
    size_y=497.0,
    size_z=130.0,
    sites=create_homogeneous_resources(
      klass=ResourceHolder,
      locations=[
        Coordinate(17.1, 36.25, 115.15),
        Coordinate(17.1, 182.25, 115.15),
        Coordinate(17.1, 328.25, 115.15),
      ],
      resource_size_x=82.6,
      resource_size_y=122.4,
      name_prefix=name,
    ),
    model="TIP_CAR_288_B00",
  )


def TIP_CAR_288_C00(name: str) -> TipCarrier:
  """Carrier for 3 Racks with 96 Tips portrait [revision C00]"""
  return TipCarrier(
    name=name,
    size_x=90.0,
    size_y=497.0,
    size_z=130.0,
    sites=create_homogeneous_resources(
      klass=ResourceHolder,
      locations=[
        Coordinate(3.7, 36.3, 114.7),
        Coordinate(3.7, 182.3, 114.7),
        Coordinate(3.7, 328.3, 114.7),
      ],
      resource_size_x=82.6,
      resource_size_y=122.4,
      name_prefix=name,
    ),
    model="TIP_CAR_288_C00",
  )


def TIP_CAR_384BC_A00(name: str) -> TipCarrier:
  """Tip carrier with 4 empty tip rack positions landscape, with Barcode Identification  [revision A00]"""
  return TipCarrier(
    name=name,
    size_x=135.0,
    size_y=497.0,
    size_z=130.0,
    sites=create_homogeneous_resources(
      klass=ResourceHolder,
      locations=[
        Coordinate(6.3, 78.2, 114.8),
        Coordinate(6.3, 163.1, 114.8),
        Coordinate(6.3, 248.1, 114.8),
        Coordinate(6.3, 333.1, 114.8),
      ],
      resource_size_x=122.4,
      resource_size_y=82.6,
      name_prefix=name,
    ),
    model="TIP_CAR_384BC_A00",
  )


def TIP_CAR_384_A00(name: str) -> TipCarrier:
  """Carrier for 4 Racks with 96 Tips landscape [revision A00]"""
  return TipCarrier(
    name=name,
    size_x=135.0,
    size_y=497.0,
    size_z=130.0,
    sites=create_homogeneous_resources(
      klass=ResourceHolder,
      locations=[
        Coordinate(6.3, 78.2, 114.8),
        Coordinate(6.3, 163.1, 114.8),
        Coordinate(6.3, 248.1, 114.8),
        Coordinate(6.3, 333.1, 114.8),
      ],
      resource_size_x=122.4,
      resource_size_y=82.6,
      name_prefix=name,
    ),
    model="TIP_CAR_384_A00",
  )


def TIP_CAR_480(name: str) -> TipCarrier:
  """Carrier for 5 Racks with 96 Tips landscape [revision A00]"""
  return TipCarrier(
    name=name,
    size_x=135.0,
    size_y=497.0,
    size_z=130.0,
    sites=create_homogeneous_resources(
      klass=ResourceHolder,
      locations=[
        Coordinate(6.2, 10.0, 114.95),
        Coordinate(6.2, 106.0, 114.95),
        Coordinate(6.2, 202.0, 114.95),
        Coordinate(6.2, 298.0, 114.95),
        Coordinate(6.2, 394.0, 114.95),
      ],
      resource_size_x=122.4,
      resource_size_y=82.6,
      name_prefix=name,
    ),
    model="TIP_CAR_480",
  )


def TIP_CAR_480BC_A00(name: str) -> TipCarrier:
  """Tip carrier with 5 tip rack positions landscape"""
  return TipCarrier(
    name=name,
    size_x=135.0,
    size_y=497.0,
    size_z=130.0,
    sites=create_homogeneous_resources(
      klass=ResourceHolder,
      locations=[
        Coordinate(6.2, 10.0, 114.95),
        Coordinate(6.2, 106.0, 114.95),
        Coordinate(6.2, 202.0, 114.95),
        Coordinate(6.2, 298.0, 114.95),
        Coordinate(6.2, 394.0, 114.95),
      ],
      resource_size_x=122.4,
      resource_size_y=82.6,
      name_prefix=name,
    ),
    model="TIP_CAR_480BC_A00",
  )


def TIP_CAR_480_A00(name: str) -> TipCarrier:
  """Carrier for 5 Racks with 96 Tips landscape [revision A00]"""
  return TipCarrier(
    name=name,
    size_x=135.0,
    size_y=497.0,
    size_z=130.0,
    sites=create_homogeneous_resources(
      klass=ResourceHolder,
      locations=[
        Coordinate(6.2, 10.0, 114.95),
        Coordinate(6.2, 106.0, 114.95),
        Coordinate(6.2, 202.0, 114.95),
        Coordinate(6.2, 298.0, 114.95),
        Coordinate(6.2, 394.0, 114.95),
      ],
      resource_size_x=122.4,
      resource_size_y=82.6,
      name_prefix=name,
    ),
    model="TIP_CAR_480_A00",
  )


def TIP_CAR_72_4mlTF_C00(name: str) -> TipCarrier:
  """Tip carrier with 3 4ml tip with filter racks portrait  [revision C00]"""
  return TipCarrier(
    name=name,
    size_x=90.0,
    size_y=497.0,
    size_z=130.0,
    sites=create_homogeneous_resources(
      klass=ResourceHolder,
      locations=[
        Coordinate(3.7, 36.3, 114.7),
        Coordinate(3.7, 182.3, 114.7),
        Coordinate(3.7, 328.3, 114.7),
      ],
      resource_size_x=82.6,
      resource_size_y=122.4,
      name_prefix=name,
    ),
    model="TIP_CAR_72_4mlTF_C00",
  )


def TIP_CAR_72_5mlT_C00(name: str) -> TipCarrier:
  """Tip carrier with 3 5ml tip racks portrait  [revision C00]"""
  return TipCarrier(
    name=name,
    size_x=90.0,
    size_y=497.0,
    size_z=130.0,
    sites=create_homogeneous_resources(
      klass=ResourceHolder,
      locations=[
        Coordinate(3.7, 36.3, 114.7),
        Coordinate(3.7, 182.3, 114.7),
        Coordinate(3.7, 328.3, 114.7),
      ],
      resource_size_x=82.6,
      resource_size_y=122.4,
      name_prefix=name,
    ),
    model="TIP_CAR_72_5mlT_C00",
  )


def TIP_CAR_96BC_4mlTF_A00(name: str) -> TipCarrier:
  """Carrier for 4 4ml with filter tip racks landscape"""
  return TipCarrier(
    name=name,
    size_x=135.0,
    size_y=497.0,
    size_z=130.0,
    sites=create_homogeneous_resources(
      klass=ResourceHolder,
      locations=[
        Coordinate(6.3, 78.2, 114.8),
        Coordinate(6.3, 163.1, 114.8),
        Coordinate(6.3, 248.1, 114.8),
        Coordinate(6.3, 333.1, 114.8),
      ],
      resource_size_x=122.4,
      resource_size_y=82.6,
      name_prefix=name,
    ),
    model="TIP_CAR_96BC_4mlTF_A00",
  )


def TIP_CAR_96BC_5mlT_A00(name: str) -> TipCarrier:
  """Carrier for 4 5ml tip racks landscape"""
  return TipCarrier(
    name=name,
    size_x=135.0,
    size_y=497.0,
    size_z=130.0,
    sites=create_homogeneous_resources(
      klass=ResourceHolder,
      locations=[
        Coordinate(6.3, 78.2, 114.8),
        Coordinate(6.3, 163.1, 114.8),
        Coordinate(6.3, 248.1, 114.8),
        Coordinate(6.3, 333.1, 114.8),
      ],
      resource_size_x=122.4,
      resource_size_y=82.6,
      name_prefix=name,
    ),
    model="TIP_CAR_96BC_5mlT_A00",
  )


def TIP_CAR_NTR_A00(name: str) -> TipCarrier:
  """Carrier with 5 nestable tip rack positions"""
  return TipCarrier(
    name=name,
    size_x=135.0,
    size_y=497.0,
    size_z=130.0,
    sites=create_homogeneous_resources(
      klass=ResourceHolder,
      locations=[
        Coordinate(6.2, 10.0, 29.0),
        Coordinate(6.2, 106.0, 29.0),
        Coordinate(6.2, 202.0, 29.0),
        Coordinate(6.2, 298.0, 29.0),
        Coordinate(6.2, 394.0, 29.0),
      ],
      resource_size_x=122.4,
      resource_size_y=82.6,
      name_prefix=name,
    ),
    model="TIP_CAR_NTR_A00",
  )



from .backend import MolecularDevicesBackend


class MolecularDevicesSpectraMaxM5Backend(MolecularDevicesBackend):
  """Backend for Molecular Devices SpectraMax M5 plate readers."""

  def __init__(self, port: str) -> None:
    super().__init__(port)



from abc import ABCMeta, abstractmethod
from typing import List

from pylabrobot.machines.backend import MachineBackend


class PumpBackend(MachineBackend, metaclass=ABCMeta):
  """Abstract base class for pump backends."""

  def run_revolutions(self, num_revolutions: float):
    """Run for a given number of revolutions.

    Args:
      num_revolutions: number of revolutions to run.
    """

  def run_continuously(self, speed: float):
    """Run continuously at a given speed.

    If speed is 0, the pump will be halted.

    Args:
      speed: speed in rpm/pump-specific units.
    """

  def halt(self):
    """Halt the pump."""

  async def stop(self):
    """Close the connection to the pump."""


class PumpArrayBackend(MachineBackend, metaclass=ABCMeta):
  """
  Abstract base class for pump array backends.

  For more information on some methods and arguments, see the documentation for the
  :class:`~PumpArray` class.
  """

  @property
  @abstractmethod
  def num_channels(self) -> int:
    """The number of channels that the pump array has."""

  async def run_revolutions(self, num_revolutions: List[float], use_channels: List[int]):
    """Run the specified channels at the speed selected.
    If speed is 0, the pump will be halted.

    Args:
      num_revolutions: number of revolutions to run pumps.
      use_channels: pump array channels to run
    """

  async def run_continuously(self, speed: List[float], use_channels: List[int]):
    """Run for a given number of revolutions.
    Args:
      speed: rate at which to run pump.
      use_channels: pump array channels to run
    """

  async def halt(self):
    """Halt the entire pump array."""

  async def stop(self):
    """Close the connection to the pump array."""



from __future__ import annotations

import asyncio
from dataclasses import dataclass
from enum import Enum, auto
from itertools import count
from typing import Any, Awaitable, Callable, Optional


class OpState(Enum):
    PENDING = auto()
    RUNNING = auto()
    DONE = auto()
    FAILED = auto()
    CANCELLED = auto()


@dataclass
class OpInfo:
    state: OpState
    future: asyncio.Future
    result: Any = None
    error: Optional[BaseException] = None


class AsyncRW:
    """
    - Two queues: read_q, write_q
    - Two worker tasks created at start()
    - read()/write() enqueue and return op_id
    - poll status via status(op_id) / done(op_id) / result(op_id)
    """

    def __init__(
        self,
        read_handler: Callable[[Any], Awaitable[Any]],
        write_handler: Callable[[Any], Awaitable[Any]],
        *,
        max_queue: int = 0,
    ) -> None:
        self._read_handler = read_handler
        self._write_handler = write_handler

        self._read_q: asyncio.Queue[tuple[str, Any]] = asyncio.Queue(maxsize=max_queue)
        self._write_q: asyncio.Queue[tuple[str, Any]] = asyncio.Queue(maxsize=max_queue)

        self._ops: dict[str, OpInfo] = {}
        self._id_read = count(1)
        self._id_write = count(1)

        self._stop = asyncio.Event()
        self._read_worker: Optional[asyncio.Task] = None
        self._write_worker: Optional[asyncio.Task] = None

    # --- lifecycle ---------------------------------------------------------

    async def start(self) -> None:
        if self._read_worker or self._write_worker:
            return
        self._stop.clear()
        self._read_worker = asyncio.create_task(self._run_read_worker(), name="rw:read-worker")
        self._write_worker = asyncio.create_task(self._run_write_worker(), name="rw:write-worker")

    async def close(self) -> None:
        self._stop.set()
        # Cancel workers so they exit promptly even if blocked on .get()
        for t in (self._read_worker, self._write_worker):
            if t:
                t.cancel()
        await asyncio.gather(
            *(t for t in (self._read_worker, self._write_worker) if t),
            return_exceptions=True,
        )
        self._read_worker = None
        self._write_worker = None

    @property
    def read_worker_id(self) -> Optional[str]:
        return self._read_worker.get_name() if self._read_worker else None

    @property
    def write_worker_id(self) -> Optional[str]:
        return self._write_worker.get_name() if self._write_worker else None

    # --- enqueue ops -------------------------------------------------------

    def read(self, req: Any) -> str:
        op_id = f"r{next(self._id_read)}"
        fut = asyncio.get_running_loop().create_future()
        self._ops[op_id] = OpInfo(state=OpState.PENDING, future=fut)
        self._read_q.put_nowait((op_id, req))
        return op_id

    def write(self, item: Any) -> str:
        op_id = f"w{next(self._id_write)}"
        fut = asyncio.get_running_loop().create_future()
        self._ops[op_id] = OpInfo(state=OpState.PENDING, future=fut)
        self._write_q.put_nowait((op_id, item))
        return op_id

    # --- polling / results -------------------------------------------------

    def status(self, op_id: str) -> OpState:
        return self._ops[op_id].state

    def done(self, op_id: str) -> bool:
        return self._ops[op_id].future.done()

    async def wait(self, op_id: str) -> Any:
        return await self._ops[op_id].future

    def result(self, op_id: str) -> Any:
        # raises if not done / failed, like Future.result()
        return self._ops[op_id].future.result()

    def exception(self, op_id: str) -> Optional[BaseException]:
        return self._ops[op_id].future.exception()

    def cancel(self, op_id: str) -> bool:
        info = self._ops.get(op_id)
        if not info:
            return False
        info.state = OpState.CANCELLED
        return info.future.cancel()

    # --- workers -----------------------------------------------------------

    async def _run_read_worker(self) -> None:
        try:
            while not self._stop.is_set():
                op_id, req = await self._read_q.get()
                await self._execute(op_id, self._read_handler, req)
                self._read_q.task_done()
        except asyncio.CancelledError:
            pass

    async def _run_write_worker(self) -> None:
        try:
            while not self._stop.is_set():
                op_id, item = await self._write_q.get()
                await self._execute(op_id, self._write_handler, item)
                self._write_q.task_done()
        except asyncio.CancelledError:
            pass

    async def _execute(
        self,
        op_id: str,
        handler: Callable[[Any], Awaitable[Any]],
        payload: Any,
    ) -> None:
        info = self._ops.get(op_id)
        if not info or info.future.cancelled():
            return

        info.state = OpState.RUNNING
        try:
            res = await handler(payload)
            info.result = res
            info.state = OpState.DONE
            if not info.future.done():
                info.future.set_result(res)
        except asyncio.CancelledError:
            info.state = OpState.CANCELLED
            if not info.future.done():
                info.future.cancel()
        except BaseException as e:
            info.error = e
            info.state = OpState.FAILED
            if not info.future.done():
                info.future.set_exception(e)


"""EL406 shake/soak step methods.

Provides the shake operation and its command builder.
"""

from __future__ import annotations

import logging
from typing import Literal

from pylabrobot.io.binary import Writer
from pylabrobot.resources import Plate

from ..helpers import plate_to_wire_byte
from ..protocol import build_framed_message
from ._base import EL406StepsBaseMixin

INTENSITY_TO_BYTE: dict[str, int] = {
  "Variable": 0x01,
  "Slow": 0x02,
  "Medium": 0x03,
  "Fast": 0x04,
}

logger = logging.getLogger(__name__)


Intensity = Literal["Variable", "Slow", "Medium", "Fast"]


def validate_intensity(intensity: Intensity) -> None:
  if intensity not in {"Slow", "Medium", "Fast", "Variable"}:
    raise ValueError(
      f"intensity must be one of {sorted({'Slow', 'Medium', 'Fast', 'Variable'})}, got {intensity!r}"
    )


class EL406ShakeStepsMixin(EL406StepsBaseMixin):
  """Mixin for shake/soak step operations."""

  MAX_SHAKE_DURATION = 3599  # 59:59 max (mm:ss format, mm max=59)
  MAX_SOAK_DURATION = 3599  # 59:59 max (mm:ss format, mm max=59)

  async def shake(
    self,
    plate: Plate,
    duration: int = 0,
    intensity: Intensity = "Medium",
    soak_duration: int = 0,
    move_home_first: bool = True,
  ) -> None:
    """Shake the plate with optional soak period.

    Durations are in whole seconds (GUI uses mm:ss picker, max 59:59 each).
    A duration of 0 disables shake. A soak_duration of 0 disables soak.

    Note: The GUI forces move_home_first=True when total time exceeds 60s
    to prevent manifold drip contamination. Our default of True matches this.

    Args:
      plate: PLR Plate resource.
      duration: Shake duration in seconds (0-3599). 0 to disable shake.
      intensity: Shake intensity - "Variable", "Slow" (3.5 Hz),
                 "Medium" (5 Hz), or "Fast" (8 Hz).
      soak_duration: Soak duration in seconds after shaking (0-3599). 0 to disable.
      move_home_first: Move carrier to home position before shaking (default True).

    Raises:
      ValueError: If parameters are invalid.
    """
    if duration < 0 or duration > self.MAX_SHAKE_DURATION:
      raise ValueError(f"Invalid duration {duration}. Must be 0-{self.MAX_SHAKE_DURATION}.")
    if soak_duration < 0 or soak_duration > self.MAX_SOAK_DURATION:
      raise ValueError(
        f"Invalid soak_duration {soak_duration}. Must be 0-{self.MAX_SOAK_DURATION}."
      )
    if duration == 0 and soak_duration == 0:
      raise ValueError("At least one of duration or soak_duration must be > 0.")
    validate_intensity(intensity)

    shake_enabled = duration > 0

    logger.info(
      "Shake: %ds, %s intensity, move_home=%s, soak=%ds",
      duration,
      intensity,
      move_home_first,
      soak_duration,
    )

    data = self._build_shake_command(
      plate=plate,
      shake_duration=duration,
      soak_duration=soak_duration,
      intensity=intensity,
      shake_enabled=shake_enabled,
      move_home_first=move_home_first,
    )
    framed_command = build_framed_message(command=0xA3, data=data)
    total_timeout = duration + soak_duration + self.timeout
    async with self.batch(plate):
      await self._send_step_command(framed_command, timeout=total_timeout)

  # =========================================================================
  # COMMAND BUILDERS
  # =========================================================================

  def _build_shake_command(
    self,
    plate: Plate,
    shake_duration: int = 0,
    soak_duration: int = 0,
    intensity: Intensity = "Medium",
    shake_enabled: bool = True,
    move_home_first: bool = True,
  ) -> bytes:
    """Build shake command bytes.

    Byte structure (12 bytes):
      [0]      Plate type
      [1]      move_home_first: 0x00 or 0x01
      [2-3]    Shake duration in total seconds (16-bit LE)
      [4]      Intensity: 0x01=Variable, 0x02=Slow, 0x03=Medium, 0x04=Fast
      [5]      Reserved: 0x00
      [6-7]    Soak duration in total seconds (16-bit LE)
      [8-11]   Padding (4 bytes)

    Args:
      plate: PLR Plate resource.
      shake_duration: Shake duration in seconds.
      soak_duration: Soak duration in seconds.
      intensity: Shake intensity ("Variable", "Slow", "Medium", "Fast").
      shake_enabled: Whether shake is enabled. When False, shake_duration is not encoded.
      move_home_first: Move carrier to home position before shaking (default True).

    Returns:
      Command bytes (12 bytes).
    """
    shake_total_seconds = int(shake_duration) if shake_enabled else 0

    return (
      Writer()
      .u8(plate_to_wire_byte(plate))                    # [0] Plate type
      .u8(0x01 if move_home_first else 0x00)           # [1] move_home_first
      .u16(shake_total_seconds)                        # [2-3] Shake duration (seconds)
      .u8(INTENSITY_TO_BYTE.get(intensity, 0x03))      # [4] Intensity
      .u8(0x00)                                        # [5] Reserved
      .u16(int(soak_duration))                         # [6-7] Soak duration (seconds)
      .raw_bytes(b'\x00' * 4)                          # [8-11] Padding
      .finish()
    )  # fmt: skip



from .container import Container


class Trash(Container):
  """Trash area."""

  def __init__(
    self,
    name,
    size_x,
    size_y,
    size_z,
    material_z_thickness=0,
    max_volume=float("inf"),
    category="trash",
    model=None,
    compute_volume_from_height=None,
    compute_height_from_volume=None,
    height_volume_data=None,
  ):
    super().__init__(
      name=name,
      size_x=size_x,
      size_y=size_y,
      size_z=size_z,
      material_z_thickness=material_z_thickness,
      max_volume=max_volume,
      category=category,
      model=model,
      compute_volume_from_height=compute_volume_from_height,
      compute_height_from_volume=compute_height_from_volume,
      height_volume_data=height_volume_data,
    )




  async def binary_interpreter(
    self,
    node_id: int,
    cmd: str,
    cmd_index: int,
    cmd_type: eCmdType,
    value_ref: str,  # Simulates ref string
    val_type: eValType = eValType.Int,
    response_time: int = 0,  # Simulates ref long
    low_priority: bool = False,
  ) -> int:
    """
    Interprets and sends binary commands to the motor.

    This is a critical and complex method that involves:
    1. Formatting command and data into bytes.
    2. Sending bytes via CAN_Write.
    3. Waiting for a response (polling `query_wait_buffer`).
    4. Parsing the response based on CmdType and ValType.
    """

    if self.grp_node_id is None:  # C# `if (this.GrpNodeID == null)`
      return 0  # Or handle appropriately

    # C# dynamic array `numArray1` (int[8]) used for msg bytes
    # C# `flag1` is for ValQuery, `flag2` for Float types, `Execute` for Execute CmdType
    is_val_query = cmd_type == eCmdType.ValQuery
    is_float_type = val_type == eValType.Float
    is_execute_cmd = cmd_type == eCmdType.Execute

    query_idx = 0
    group_query_indices: List[int] = []  # For NodeID 10 (broadcast group query)

    if node_id != 10:  # C# MAGIC NUMBER for Group broadcast (0x0A)
      query_idx = self.get_next_query_wait_buffer_index(node_id)
    else:
      if self.grp_node_id:
        for grp_node in self.grp_node_id:
          group_query_indices.append(self.get_next_query_wait_buffer_index(grp_node))

    for retry in range(5):
      if not value_ref:  # C# `Operators.CompareString(Value, "", false) == 0`
        value_ref = "0"

      # Prepare QueryWaitBuffer entry
      if node_id != 10:
        node_idx = self.get_node_id_index(node_id)
        self.query_wait_buffer[node_idx][query_idx].node_id = node_id
        self.query_wait_buffer[node_idx][query_idx].msg_index = cmd_index
        self.query_wait_buffer[node_idx][query_idx].msg_type = cmd
        # self.query_wait_buffer[node_idx][query_idx].response = ""
      else:
        if self.grp_node_id:
          for i, grp_node in enumerate(self.grp_node_id):
            node_idx = self.get_node_id_index(grp_node)
            current_query_idx = group_query_indices[i]
            self.query_wait_buffer[node_idx][current_query_idx].node_id = grp_node
            self.query_wait_buffer[node_idx][current_query_idx].msg_index = cmd_index
            self.query_wait_buffer[node_idx][current_query_idx].msg_type = cmd
            # self.query_wait_buffer[node_idx][current_query_idx].response = ""

      # Construct CAN message data bytes (8 bytes total for RPDO2)
      # C# `numArray1` (int[8]) used for msg bytes
      can_data: List[int] = [0] * 8
      can_data[0] = ord(cmd[:1])
      can_data[1] = ord(vb_right(cmd, 1))

      binary_cmd_index = int_to_binary(cmd_index, 14)  # C# fixed length 14
      can_data[2] = binary_to_int(vb_right(binary_cmd_index, 8))  # Lower 8 bits

      # Upper 6 bits of cmd_index, plus flags for query (bit 6) and float (bit 7)
      upper_6_bits_cmd_idx = binary_to_int(binary_cmd_index[:6])

      # C# uses Math.Pow(2.0, checked(2 * Math.Abs(-(flag1 ? 1 : 0))), 6.0)))
      # If flag1 is true (is_val_query), then 2 * 1 = 2. pow(2,6) -> bit 6 is set
      # This is bit 6 for query (0x40) and bit 7 for float (0x80)
      if is_val_query:  # Bit 6 (0x40)
        can_data[3] |= 0x40
      if is_float_type:  # Bit 7 (0x80)
        can_data[3] |= 0x80

      can_data[3] |= upper_6_bits_cmd_idx  # Merge with upper 6 bits of cmd_index (0-5)

      if is_float_type:
        # C# use BitConverter.GetBytes(Conversions.ToSingle(Value))
        # Python's struct.pack for float (little-endian)
        packed_float = struct.pack("<f", float(value_ref))
        can_data[4] = packed_float[0]
        can_data[5] = packed_float[1]
        can_data[6] = packed_float[2]
        can_data[7] = packed_float[3]
      else:
        # C# GetByte(Conversions.ToLong(Value), byte_num)
        val_long = int(value_ref)
        can_data[4] = (val_long >> 0) & 0xFF
        can_data[5] = (val_long >> 8) & 0xFF
        can_data[6] = (val_long >> 16) & 0xFF
        can_data[7] = (val_long >> 24) & 0xFF

      # Wait for any ongoing high-priority writes if low_priority is true
      if low_priority:
        start_hp_wait = time.monotonic()
        while (
          self.can_msg_buffer_first_index != self.can_msg_buffer_last_index
          and (time.monotonic() - start_hp_wait) < 10.0
        ):  # C# 10000ms
          await asyncio.sleep(0.001)

      tx_start_time = time.monotonic()
      err_code = await self.can_write(
        eCOBType.RPDO2,
        node_id,
        byte0=can_data[0],
        byte1=can_data[1],
        byte2=can_data[2],
        byte3=can_data[3],
        byte4=can_data[4],
        byte5=can_data[5],
        byte6=can_data[6],
        byte7=can_data[7],
        execute=is_execute_cmd,
        low_priority=low_priority,
      )
      if err_code != 0:
        self.sending_sv_command = False
        return err_code

      timeout_ms = 1000  # Default C# timeout
      if cmd == "SV":
        timeout_ms = 10000  # Specific timeout for SV command in C#

      timeout_sec = timeout_ms / 1000.0

      # Wait for response (different logic for single node vs. group)
      if node_id != 10:
        node_idx = self.get_node_id_index(node_id)
        response_start_time = time.monotonic()
        while (time.monotonic() - response_start_time) < timeout_sec:
          if self.query_wait_buffer[node_idx][query_idx].response != "":
            break
          await asyncio.sleep(0.001)

        response_time = int((time.monotonic() - tx_start_time) * 1000)  # milliseconds

        if not is_val_query and not is_execute_cmd:  # Value Set command, check if response matches
          is_match = False
          if (
            is_float_type
            and vb_isnumeric(value_ref)
            and vb_isnumeric(self.query_wait_buffer[node_idx][query_idx].response)
          ):
            val1 = float(value_ref)
            val2 = float(self.query_wait_buffer[node_idx][query_idx].response)
            if val1 == val2 or (val1 / val2 > 0.99 and val1 / val2 < 1.01):
              is_match = True
          elif self.query_wait_buffer[node_idx][query_idx].response.upper() == value_ref.upper():
            is_match = True

          if not is_match:
            if retry == 4:  # Last retry
              self.error_code[
                4
              ] = f"Unexpected CAN response: Attempted to send {cmd}[{cmd_index}]={value_ref}, but received response = {self.query_wait_buffer[node_idx][query_idx].response}"
              return 4
            continue  # Retry
        elif (
          self.query_wait_buffer[node_idx][query_idx].response == ""
        ):  # ValQuery or Execute, but no response
          if retry == 4:
            return 45  # Failed to receive response from motor
          continue  # Retry
        else:  # Success for ValQuery/Execute or value matched for ValSet
          break  # Exit retry loop

      else:  # Group query (NodeID == 10)
        # Need to wait for all nodes in the group to respond for *ValQuery* or *Execute*.
        # For ValSet, we need all nodes to confirm the value.
        all_responded = False
        response_start_time = time.monotonic()
        while (time.monotonic() - response_start_time) < timeout_sec:
          all_responded = True
          if self.grp_node_id:
            for i, grp_node in enumerate(self.grp_node_id):
              node_idx = self.get_node_id_index(grp_node)
              current_query_idx = group_query_indices[i]
              if self.query_wait_buffer[node_idx][current_query_idx].response == "":
                all_responded = False
                break
          if all_responded:
            break
          await asyncio.sleep(0.001)

        response_time = int((time.monotonic() - tx_start_time) * 1000)

        if not is_val_query and not is_execute_cmd:  # Value Set for group
          group_match = True
          if self.grp_node_id:
            for i, grp_node in enumerate(self.grp_node_id):
              node_idx = self.get_node_id_index(grp_node)
              current_query_idx = group_query_indices[i]
              is_match_node = False
              if (
                is_float_type
                and vb_isnumeric(value_ref)
                and vb_isnumeric(self.query_wait_buffer[node_idx][current_query_idx].response)
              ):
                val1 = float(value_ref)
                val2 = float(self.query_wait_buffer[node_idx][current_query_idx].response)
                if val1 == val2 or (val1 / val2 > 0.99 and val1 / val2 < 1.01):
                  is_match_node = True
              elif (
                self.query_wait_buffer[node_idx][current_query_idx].response.upper()
                == value_ref.upper()
              ):
                is_match_node = True

              if not is_match_node:
                group_match = False
                break
          if not group_match:
            if retry == 4:
              self.error_code[
                4
              ] = f"Unexpected CAN response (group): Attempted to send {cmd}[{cmd_index}]={value_ref}, but responses did not match."
              return 4
            continue  # Retry
          else:
            break  # Success
        elif not all_responded:  # No response from all nodes in group for ValQuery/Execute
          if retry == 4:
            return 45
          continue
        else:
          break

    if is_val_query:
      if node_id != 10:
        node_idx = self.get_node_id_index(node_id)
        value_ref = self.query_wait_buffer[node_idx][query_idx].response
      else:
        collected_responses: List[str] = []
        if self.grp_node_id:
          for i, grp_node in enumerate(self.grp_node_id):
            node_idx = self.get_node_id_index(grp_node)
            current_query_idx = group_query_indices[i]
            collected_responses.append(self.query_wait_buffer[node_idx][current_query_idx].response)
        value_ref = ",".join(collected_responses)

    self.sending_sv_command = False
    if retry == 4:  # If loop finished by exhausting retries, it means an error occurred
      return 4  # C# returned 4 for "unexpected CAN response" (already set in self.error_code)

    return 0




async def _move_thread(self):
  print("Move Thread Started")
  self.b_move_thread_running = True
  tick_count = time.monotonic()  # High-resolution timer for time differences

  while self.b_move_thread_running:
    # check the move buffer
    if self.move_buffer_first_index != self.move_buffer_last_index:
      if (time.monotonic() - tick_count) >= 1.0:  # 1000ms
        tick_count = time.monotonic()

      if self.move_buffer_first_index == 99:
        self.move_buffer_first_index = 0
      else:
        self.move_buffer_first_index += 1

      current_move = self.move_buffer[self.move_buffer_first_index]

      if current_move.pending:
        current_move.pending = False  # Mark as processed
        self.move_error_code = 0
        self.emcy_move_error_received = False

        err_code = 0
        if current_move.move_type == eMoveType.MotorMove:
          param = self.motor_move_param_buffer[current_move.index]
          err_code = await self.motor_move_execute(
            param.node_id,
            param.position,
            param.velocity,
            param.acceleration,
            current_move.index,
            param.relative,
            current_move.timeout_msec,
            param.mtr_joint_move_direction,
            current_move.wait_until_done,
          )
        elif current_move.move_type == eMoveType.MotorsMoveAbsolute:
          param = self.motors_move_absolute_param_buffer[current_move.index]
          if (
            param.node_id
            and param.position
            and param.velocity
            and param.acceleration
            and param.mtrs_joint_move_direction
          ):
            err_code = await self.motors_move_absolute_execute(
              param.node_id,
              param.position,
              param.velocity,
              param.acceleration,
              current_move.index,
              param.mtrs_joint_move_direction,
              current_move.timeout_msec,
            )
        elif current_move.move_type == eMoveType.MotorsMovePath:
          param = self.motors_move_path_param_buffer[current_move.index]
          if param.pva_point and param.stop_deceleration and param.skip_ax:
            err_code = await self.motors_move_path_execute(
              param.pva_point,
              param.time_interval_msec,
              param.stop_deceleration,
              current_move.index,
              param.skip_ax,
              current_move.timeout_msec,
            )

        current_move.error_code = err_code

      if current_move.error_code > 0:
        event_data = sEventData(
          event_type=eEventType.MoveError,
          me_error_code=current_move.error_code,
          me_index=current_move.index,
        )
        self.move_buffer_first_index = self.move_buffer_last_index  # Clear buffer on error
        for i in range(100):
          self.move_buffer[i].pending = False
        await self._raise_an_event(event_data)

    await asyncio.sleep(0.001)
  print("Move Thread Stopped")



import unittest
from unittest.mock import AsyncMock, call, patch

from pylabrobot.io.usb import USB
from pylabrobot.plate_reading.tecan.infinite_backend import (
  ExperimentalTecanInfinite200ProBackend,
  _absorbance_od_calibrated,
  _AbsorbanceRunDecoder,
  _consume_leading_ascii_frame,
  _FluorescenceRunDecoder,
  _LuminescenceRunDecoder,
)
from pylabrobot.resources import Coordinate, Plate, Well, create_ordered_items_2d
from pylabrobot.resources.tecan.plates import Plate_384_Well


def _pack_u16(words):
  return b"".join(int(word).to_bytes(2, "big") for word in words)


def _bin_blob(payload):
  payload_len = len(payload)
  trailer = b"\x00\x00\x00\x00"
  return payload_len, payload + trailer


def _abs_calibration_blob(ex_decitenth, meas_dark, meas_bright, ref_dark, ref_bright):
  header = _pack_u16([0, ex_decitenth])
  item = (0).to_bytes(4, "big") + _pack_u16([0, 0, meas_dark, meas_bright, 0, ref_dark, ref_bright])
  return _bin_blob(header + item)


def _abs_data_blob(ex_decitenth, meas, ref):
  payload = _pack_u16([0, ex_decitenth, 0, 0, 0, meas, ref])
  return _bin_blob(payload)


def _flr_calibration_blob(ex_decitenth, meas_dark, ref_dark, ref_bright):
  words = [ex_decitenth, 0, 0, 0, 0, meas_dark, 0, ref_dark, ref_bright]
  return _bin_blob(_pack_u16(words))


def _flr_data_blob(ex_decitenth, em_decitenth, meas, ref):
  words = [0, ex_decitenth, em_decitenth, 0, 0, 0, meas, ref]
  return _bin_blob(_pack_u16(words))


def _lum_data_blob(em_decitenth: int, intensity: int):
  payload = bytearray(14)
  payload[0:2] = (0).to_bytes(2, "big")
  payload[2:4] = int(em_decitenth).to_bytes(2, "big")
  payload[10:14] = int(intensity).to_bytes(4, "big", signed=True)
  return _bin_blob(bytes(payload))


def _make_test_plate():
  plate = Plate(
    "plate",
    size_x=30,
    size_y=20,
    size_z=10,
    ordered_items=create_ordered_items_2d(
      Well,
      num_items_x=3,
      num_items_y=2,
      dx=1,
      dy=2,
      dz=0,
      item_dx=10,
      item_dy=8,
      size_x=4,
      size_y=4,
      size_z=5,
    ),
  )
  plate.location = Coordinate.zero()
  return plate


def _egg_grid():
  return [
    [
      0,
      0,
      0,
      0,
      0,
      0,
      0,
      0,
      0,
      0,
      9,
      31,
      46,
      42,
      7,
      2,
      0,
      0,
      0,
      0,
      0,
      0,
      0,
      0,
    ],
    [
      0,
      0,
      0,
      0,
      0,
      0,
      0,
      0,
      1,
      24,
      69,
      100,
      137,
      142,
      70,
      24,
      0,
      0,
      0,
      0,
      0,
      0,
      0,
      0,
    ],
    [
      0,
      0,
      0,
      0,
      0,
      0,
      0,
      0,
      24,
      77,
      128,
      135,
      123,
      68,
      52,
      26,
      0,
      0,
      0,
      0,
      0,
      0,
      0,
      0,
    ],
    [
      0,
      0,
      0,
      0,
      0,
      0,
      0,
      4,
      60,
      104,
      114,
      86,
      72,
      48,
      2,
      2,
      0,
      0,
      0,
      0,
      0,
      0,
      0,
      0,
    ],
    [
      0,
      0,
      0,
      0,
      0,
      0,
      0,
      9,
      75,
      122,
      82,
      71,
      99,
      69,
      4,
      0,
      0,
      0,
      0,
      0,
      0,
      0,
      0,
      0,
    ],
    [
      0,
      0,
      0,
      0,
      0,
      0,
      0,
      3,
      64,
      132,
      148,
      61,
      75,
      137,
      86,
      17,
      0,
      0,
      0,
      0,
      0,
      0,
      0,
      0,
    ],
    [
      0,
      0,
      0,
      0,
      0,
      0,
      0,
      0,
      23,
      98,
      160,
      87,
      92,
      139,
      133,
      65,
      0,
      0,
      0,
      0,
      0,
      0,
      0,
      0,
    ],
    [
      0,
      0,
      0,
      0,
      0,
      0,
      0,
      0,
      4,
      53,
      100,
      93,
      104,
      125,
      146,
      46,
      0,
      0,
      0,
      0,
      0,
      0,
      0,
      0,
    ],
    [
      0,
      0,
      0,
      0,
      0,
      0,
      0,
      0,
      33,
      73,
      103,
      128,
      143,
      164,
      169,
      61,
      0,
      0,
      0,
      0,
      0,
      0,
      0,
      0,
    ],
    [
      0,
      0,
      0,
      0,
      0,
      0,
      0,
      4,
      60,
      93,
      113,
      90,
      107,
      124,
      137,
      118,
      7,
      0,
      0,
      0,
      0,
      0,
      0,
      0,
    ],
    [
      0,
      0,
      0,
      0,
      0,
      0,
      0,
      3,
      64,
      97,
      98,
      63,
      94,
      95,
      135,
      121,
      8,
      0,
      0,
      0,
      0,
      0,
      0,
      0,
    ],
    [
      0,
      0,
      0,
      0,
      0,
      0,
      0,
      0,
      36,
      100,
      118,
      119,
      126,
      140,
      154,
      65,
      1,
      0,
      0,
      0,
      0,
      0,
      0,
      0,
    ],
    [
      0,
      0,
      0,
      0,
      0,
      0,
      0,
      0,
      3,
      40,
      98,
      141,
      150,
      121,
      61,
      6,
      0,
      0,
      0,
      0,
      0,
      0,
      0,
      0,
    ],
    [
      0,
      0,
      0,
      0,
      0,
      0,
      0,
      0,
      0,
      0,
      8,
      75,
      88,
      12,
      0,
      0,
      0,
      0,
      0,
      0,
      0,
      0,
      0,
      0,
    ],
    [
      0,
      0,
      0,
      0,
      0,
      0,
      0,
      0,
      0,
      0,
      0,
      45,
      53,
      0,
      0,
      0,
      0,
      0,
      0,
      0,
      0,
      0,
      0,
      0,
    ],
    [
      0,
      0,
      0,
      0,
      0,
      0,
      0,
      0,
      0,
      0,
      0,
      9,
      11,
      0,
      0,
      0,
      0,
      0,
      0,
      0,
      0,
      0,
      0,
      0,
    ],
  ]


class TestTecanInfiniteDecoders(unittest.IsolatedAsyncioTestCase):
  def setUp(self):
    self.backend = ExperimentalTecanInfinite200ProBackend()
    self.plate = Plate_384_Well(name="plate")
    self.grid = _egg_grid()
    self.max_intensity = max(max(row) for row in self.grid)
    self.scan_wells = self.backend._scan_visit_order(self.plate.get_all_items(), serpentine=True)

  def _assert_matrix(self, actual, expected):
    self.assertEqual(len(actual), len(expected))
    for row_actual, row_expected in zip(actual, expected):
      self.assertEqual(len(row_actual), len(row_expected))
      for value, exp in zip(row_actual, row_expected):
        self.assertAlmostEqual(value or 0.0, exp)

  def _run_decoder_case(self, decoder, build_packet, extract_actual):
    expected_values = []
    for well in self.scan_wells:
      intensity = self.grid[well.get_row()][well.get_column()]
      payload_len, blob, expected = build_packet(intensity)
      decoder.feed_bin(payload_len, blob)
      expected_values.append(expected)
    self.assertTrue(decoder.done)
    actual_values = extract_actual(decoder)
    matrix = self.backend._format_plate_result(self.plate, self.scan_wells, actual_values)
    expected = self.backend._format_plate_result(self.plate, self.scan_wells, expected_values)
    self._assert_matrix(matrix, expected)

  def test_decode_absorbance_pattern(self):
    wavelength = 600
    reference = 10000
    max_absorbance = 1.0
    decoder = _AbsorbanceRunDecoder(len(self.scan_wells))
    cal_len, cal_blob = _abs_calibration_blob(
      wavelength * 10,
      meas_dark=0,
      meas_bright=1000,
      ref_dark=0,
      ref_bright=1000,
    )
    decoder.feed_bin(cal_len, cal_blob)
    cal = decoder.calibration
    assert cal is not None

    def build_packet(intensity):
      target = 0.0
      if self.max_intensity:
        target = (intensity / self.max_intensity) * max_absorbance
      sample = max(1, int(round(reference / (10**target))))
      payload_len, blob = _abs_data_blob(wavelength * 10, sample, reference)
      expected = _absorbance_od_calibrated(cal, [(sample, reference)])
      return payload_len, blob, expected

    def extract_actual(decoder):
      return [
        _absorbance_od_calibrated(cal, [(meas.sample, meas.reference)])
        for meas in decoder.measurements
      ]

    self._run_decoder_case(decoder, build_packet, extract_actual)

  def test_decode_fluorescence_pattern(self):
    excitation = 485
    emission = 520
    decoder = _FluorescenceRunDecoder(len(self.scan_wells))
    cal_len, cal_blob = _flr_calibration_blob(
      excitation * 10, meas_dark=0, ref_dark=0, ref_bright=1000
    )
    decoder.feed_bin(cal_len, cal_blob)

    def build_packet(intensity):
      payload_len, blob = _flr_data_blob(excitation * 10, emission * 10, intensity, 1000)
      return payload_len, blob, intensity

    def extract_actual(decoder):
      return decoder.intensities

    self._run_decoder_case(decoder, build_packet, extract_actual)

  def test_decode_luminescence_pattern(self):
    decoder = _LuminescenceRunDecoder(len(self.scan_wells))

    def build_packet(intensity):
      payload_len, blob = _lum_data_blob(0, intensity)
      return payload_len, blob, intensity

    def extract_actual(decoder):
      return [measurement.intensity for measurement in decoder.measurements]

    self._run_decoder_case(decoder, build_packet, extract_actual)


class TestTecanInfiniteScanGeometry(unittest.IsolatedAsyncioTestCase):
  def setUp(self):
    self.backend = ExperimentalTecanInfinite200ProBackend(counts_per_mm_x=1, counts_per_mm_y=1)
    self.plate = _make_test_plate()

  def test_scan_visit_order_serpentine(self):
    order = self.backend._scan_visit_order(self.plate.get_all_items(), serpentine=True)
    identifiers = [well.get_identifier() for well in order]
    self.assertEqual(identifiers, ["A1", "A2", "A3", "B3", "B2", "B1"])

  def test_scan_visit_order_linear(self):
    order = self.backend._scan_visit_order(self.plate.get_all_items(), serpentine=False)
    identifiers = [well.get_identifier() for well in order]
    self.assertEqual(identifiers, ["A1", "A2", "A3", "B1", "B2", "B3"])

  def test_scan_range_serpentine(self):
    setattr(self.backend, "_map_well_to_stage", lambda well: (well.get_column(), well.get_row()))
    row_index, row_wells = self.backend._group_by_row(self.plate.get_all_items())[0]
    start_x, end_x, count = self.backend._scan_range(row_index, row_wells, serpentine=True)
    self.assertEqual((start_x, end_x, count), (0, 2, 3))
    row_index, row_wells = self.backend._group_by_row(self.plate.get_all_items())[1]
    start_x, end_x, count = self.backend._scan_range(row_index, row_wells, serpentine=True)
    self.assertEqual((start_x, end_x, count), (2, 0, 3))

  def test_map_well_to_stage(self):
    stage_x, stage_y = self.backend._map_well_to_stage(self.plate.get_well("A1"))
    self.assertEqual((stage_x, stage_y), (3, 8))
    stage_x, stage_y = self.backend._map_well_to_stage(self.plate.get_well("B1"))
    self.assertEqual((stage_x, stage_y), (3, 16))


class TestTecanInfiniteAscii(unittest.IsolatedAsyncioTestCase):
  def test_frame_command(self):
    framed = ExperimentalTecanInfinite200ProBackend._frame_command("A")
    self.assertEqual(framed, b"\x02A\x03\x00\x00\x01\x40\x0d")

  def test_consume_leading_ascii_frame(self):
    buffer = bytearray(ExperimentalTecanInfinite200ProBackend._frame_command("ST") + b"XYZ")
    consumed, text = _consume_leading_ascii_frame(buffer)
    self.assertTrue(consumed)
    self.assertEqual(text, "ST")
    self.assertEqual(buffer, bytearray(b"XYZ"))

  def test_terminal_frames(self):
    self.assertTrue(ExperimentalTecanInfinite200ProBackend._is_terminal_frame("ST"))
    self.assertTrue(ExperimentalTecanInfinite200ProBackend._is_terminal_frame("+"))
    self.assertTrue(ExperimentalTecanInfinite200ProBackend._is_terminal_frame("-"))
    self.assertTrue(ExperimentalTecanInfinite200ProBackend._is_terminal_frame("BY#T5000"))
    self.assertFalse(ExperimentalTecanInfinite200ProBackend._is_terminal_frame("OK"))


class TestTecanInfiniteCommands(unittest.IsolatedAsyncioTestCase):
  """Tests that verify correct commands are sent to the device."""

  def setUp(self):
    self.mock_usb = AsyncMock(spec=USB)
    self.mock_usb.setup = AsyncMock()
    self.mock_usb.stop = AsyncMock()
    self.mock_usb.write = AsyncMock()
    # Default to returning terminal response
    self.mock_usb.read = AsyncMock(return_value=self._frame("ST"))

    patcher = patch(
      "pylabrobot.plate_reading.tecan.infinite_backend.USB",
      return_value=self.mock_usb,
    )
    self.mock_usb_class = patcher.start()
    self.addCleanup(patcher.stop)

    self.backend = ExperimentalTecanInfinite200ProBackend(
      counts_per_mm_x=1000, counts_per_mm_y=1000
    )
    self.plate = _make_test_plate()
    self.plate.location = Coordinate.zero()

  def _frame(self, command: str) -> bytes:
    """Helper to frame a command."""
    return ExperimentalTecanInfinite200ProBackend._frame_command(command)

  async def test_open(self):
    self.backend._ready = True

    await self.backend.open()

    self.mock_usb.write.assert_has_calls(
      [
        call(self._frame("ABSOLUTE MTP,OUT")),
        call(self._frame("BY#T5000")),
      ]
    )

  async def test_close(self):
    self.backend._ready = True

    await self.backend.close(self.plate)

    self.mock_usb.write.assert_has_calls(
      [
        call(self._frame("ABSOLUTE MTP,IN")),
        call(self._frame("BY#T5000")),
      ]
    )

  async def test_read_absorbance_commands(self):
    """Test that read_absorbance sends the correct configuration commands."""
    self.backend._ready = True

    async def mock_await(decoder, row_count, mode):
      cal_len, cal_blob = _abs_calibration_blob(6000, 0, 1000, 0, 1000)
      decoder.feed_bin(cal_len, cal_blob)
      for _ in range(row_count):
        data_len, data_blob = _abs_data_blob(6000, 500, 1000)
        decoder.feed_bin(data_len, data_blob)

    with patch.object(self.backend, "_await_measurements", side_effect=mock_await):
      with patch.object(self.backend, "_await_scan_terminal", new_callable=AsyncMock):
        await self.backend.read_absorbance(self.plate, [], wavelength=600)

    self.mock_usb.write.assert_has_calls(
      [
        # _begin_run
        call(self._frame("KEYLOCK ON")),
        # _configure_absorbance
        call(self._frame("MODE ABS")),
        call(self._frame("EXCITATION CLEAR")),
        call(self._frame("TIME CLEAR")),
        call(self._frame("GAIN CLEAR")),
        call(self._frame("READS CLEAR")),
        call(self._frame("POSITION CLEAR")),
        call(self._frame("MIRROR CLEAR")),
        call(self._frame("EXCITATION 0,ABS,6000,90,0")),
        call(self._frame("EXCITATION 1,ABS,6000,90,0")),
        call(self._frame("READS 0,NUMBER=25")),
        call(self._frame("READS 1,NUMBER=25")),
        call(self._frame("TIME 0,READDELAY=0")),
        call(self._frame("TIME 1,READDELAY=0")),
        call(self._frame("SCAN DIRECTION=ALTUP")),
        call(self._frame("#RATIO LABELS")),
        call(self._frame("BEAM DIAMETER=700")),
        call(self._frame("RATIO LABELS=1")),
        call(self._frame("PREPARE REF")),
        # row scans (2 rows in test plate)
        call(self._frame("ABSOLUTE MTP,Y=8000")),
        call(self._frame("SCAN DIRECTION=ALTUP")),
        call(self._frame("SCANX 3000,23000,3")),
        call(self._frame("ABSOLUTE MTP,Y=16000")),
        call(self._frame("SCAN DIRECTION=ALTUP")),
        call(self._frame("SCANX 23000,3000,3")),
        # _end_run
        call(self._frame("TERMINATE")),
        call(self._frame("CHECK MTP.STEPLOSS")),
        call(self._frame("CHECK ABS.STEPLOSS")),
        call(self._frame("KEYLOCK OFF")),
        call(self._frame("ABSOLUTE MTP,IN")),
      ]
    )

  async def test_read_absorbance_uses_late_pending_calibration(self):
    self.backend._ready = True
    terminal_calls = 0

    async def mock_await(decoder, row_count, mode):
      for _ in range(row_count):
        data_len, data_blob = _abs_data_blob(6000, 500, 1000)
        decoder.feed_bin(data_len, data_blob)

    async def mock_terminal(_saw_terminal):
      nonlocal terminal_calls
      terminal_calls += 1
      if terminal_calls == 2:
        cal_len, cal_blob = _abs_calibration_blob(6000, 0, 1000, 0, 1000)
        self.backend._pending_bin_events.append((cal_len, cal_blob))

    with patch.object(self.backend, "_await_measurements", side_effect=mock_await):
      with patch.object(self.backend, "_await_scan_terminal", side_effect=mock_terminal):
        result = await self.backend.read_absorbance(self.plate, [], wavelength=600)

    self.assertAlmostEqual(result[0]["data"][0][0], 0.3010299956639812)

  async def test_read_fluorescence_commands(self):
    """Test that read_fluorescence sends the correct configuration commands."""
    self.backend._ready = True

    async def mock_await(decoder, row_count, mode):
      cal_len, cal_blob = _flr_calibration_blob(4850, 0, 0, 1000)
      decoder.feed_bin(cal_len, cal_blob)
      for _ in range(row_count):
        data_len, data_blob = _flr_data_blob(4850, 5200, 500, 1000)
        decoder.feed_bin(data_len, data_blob)

    with patch.object(self.backend, "_await_measurements", side_effect=mock_await):
      with patch.object(self.backend, "_await_scan_terminal", new_callable=AsyncMock):
        await self.backend.read_fluorescence(
          self.plate, [], excitation_wavelength=485, emission_wavelength=520
        )

    # Fluorescence config is sent twice (UI behavior)
    fl_config_commands = [
      call(self._frame("MODE FI.TOP")),
      call(self._frame("EXCITATION CLEAR")),
      call(self._frame("EMISSION CLEAR")),
      call(self._frame("TIME CLEAR")),
      call(self._frame("GAIN CLEAR")),
      call(self._frame("READS CLEAR")),
      call(self._frame("POSITION CLEAR")),
      call(self._frame("MIRROR CLEAR")),
      call(self._frame("EXCITATION 0,FI,4850,50,0")),
      call(self._frame("EMISSION 0,FI,5200,200,0")),
      call(self._frame("TIME 0,INTEGRATION=20")),
      call(self._frame("TIME 0,LAG=0")),
      call(self._frame("TIME 0,READDELAY=0")),
      call(self._frame("GAIN 0,VALUE=100")),
      call(self._frame("POSITION 0,Z=20000")),
      call(self._frame("BEAM DIAMETER=3000")),
      call(self._frame("SCAN DIRECTION=UP")),
      call(self._frame("RATIO LABELS=1")),
      call(self._frame("READS 0,NUMBER=25")),
      call(self._frame("EXCITATION 1,FI,4850,50,0")),
      call(self._frame("EMISSION 1,FI,5200,200,0")),
      call(self._frame("TIME 1,INTEGRATION=20")),
      call(self._frame("TIME 1,LAG=0")),
      call(self._frame("TIME 1,READDELAY=0")),
      call(self._frame("GAIN 1,VALUE=100")),
      call(self._frame("POSITION 1,Z=20000")),
      call(self._frame("READS 1,NUMBER=25")),
    ]

    self.mock_usb.write.assert_has_calls(
      [
        # _begin_run
        call(self._frame("KEYLOCK ON")),
        # _configure_fluorescence (sent twice)
        *fl_config_commands,
        *fl_config_commands,
        call(self._frame("PREPARE REF")),
        # row scans (2 rows in test plate)
        call(self._frame("ABSOLUTE MTP,Y=8000")),
        call(self._frame("SCAN DIRECTION=UP")),
        call(self._frame("SCANX 3000,23000,3")),
        call(self._frame("ABSOLUTE MTP,Y=16000")),
        call(self._frame("SCAN DIRECTION=UP")),
        call(self._frame("SCANX 23000,3000,3")),
        # _end_run
        call(self._frame("TERMINATE")),
        call(self._frame("CHECK MTP.STEPLOSS")),
        call(self._frame("CHECK FI.TOP.STEPLOSS")),
        call(self._frame("CHECK FI.STEPLOSS.Z")),
        call(self._frame("KEYLOCK OFF")),
        call(self._frame("ABSOLUTE MTP,IN")),
      ]
    )

  async def test_read_luminescence_commands(self):
    """Test that read_luminescence sends the correct configuration commands."""
    self.backend._ready = True

    async def mock_await(decoder, row_count, mode):
      cal_blob = bytes(14)
      decoder.feed_bin(10, cal_blob)
      for _ in range(row_count):
        data_len, data_blob = _lum_data_blob(0, 1000)
        decoder.feed_bin(data_len, data_blob)

    with patch.object(self.backend, "_await_measurements", side_effect=mock_await):
      with patch.object(self.backend, "_await_scan_terminal", new_callable=AsyncMock):
        await self.backend.read_luminescence(self.plate, [], focal_height=14.62)

    self.mock_usb.write.assert_has_calls(
      [
        # _begin_run
        call(self._frame("KEYLOCK ON")),
        # _configure_luminescence
        call(self._frame("MODE LUM")),
        call(self._frame("CHECK LUM.FIBER")),
        call(self._frame("CHECK LUM.LID")),
        call(self._frame("CHECK LUM.STEPLOSS")),
        call(self._frame("MODE LUM")),
        call(self._frame("EMISSION CLEAR")),
        call(self._frame("TIME CLEAR")),
        call(self._frame("GAIN CLEAR")),
        call(self._frame("READS CLEAR")),
        call(self._frame("POSITION CLEAR")),
        call(self._frame("MIRROR CLEAR")),
        call(self._frame("POSITION LUM,Z=14620")),
        call(self._frame("TIME 0,INTEGRATION=3000000")),
        call(self._frame("READS 0,NUMBER=25")),
        call(self._frame("SCAN DIRECTION=UP")),
        call(self._frame("RATIO LABELS=1")),
        call(self._frame("EMISSION 1,EMPTY,0,0,0")),
        call(self._frame("TIME 1,INTEGRATION=1000000")),
        call(self._frame("TIME 1,READDELAY=0")),
        call(self._frame("READS 1,NUMBER=25")),
        call(self._frame("#EMISSION ATTENUATION")),
        call(self._frame("PREPARE REF")),
        # row scans (2 rows, non-serpentine so both scan left-to-right)
        call(self._frame("ABSOLUTE MTP,Y=8000")),
        call(self._frame("SCAN DIRECTION=UP")),
        call(self._frame("SCANX 3000,23000,3")),
        call(self._frame("ABSOLUTE MTP,Y=16000")),
        call(self._frame("SCAN DIRECTION=UP")),
        call(self._frame("SCANX 3000,23000,3")),
        # _end_run
        call(self._frame("TERMINATE")),
        call(self._frame("CHECK MTP.STEPLOSS")),
        call(self._frame("CHECK LUM.STEPLOSS")),
        call(self._frame("KEYLOCK OFF")),
        call(self._frame("ABSOLUTE MTP,IN")),
      ]
    )

  async def test_read_luminescence_defaults_focal_height_to_20mm(self):
    """Test that read_luminescence defaults focal height to 20 mm."""
    self.backend._ready = True

    async def mock_await(decoder, row_count, mode):
      cal_blob = bytes(14)
      decoder.feed_bin(10, cal_blob)
      for _ in range(row_count):
        data_len, data_blob = _lum_data_blob(0, 1000)
        decoder.feed_bin(data_len, data_blob)

    with patch.object(self.backend, "_await_measurements", side_effect=mock_await):
      with patch.object(self.backend, "_await_scan_terminal", new_callable=AsyncMock):
        await self.backend.read_luminescence(self.plate, [])

    self.mock_usb.write.assert_any_call(self._frame("POSITION LUM,Z=20000"))



import asyncio
import logging
import time

try:
  import serial

  HAS_SERIAL = True
except ImportError as e:
  HAS_SERIAL = False
  _SERIAL_IMPORT_ERROR = e

from pylabrobot.barcode_scanners.backend import (
  BarcodeScannerBackend,
  BarcodeScannerError,
)
from pylabrobot.io.serial import Serial
from pylabrobot.resources.barcode import Barcode

logger = logging.getLogger(__name__)


class KeyenceBarcodeScannerBackend(BarcodeScannerBackend):
  default_baudrate = 9600
  serial_messaging_encoding = "ascii"
  init_timeout = 1.0  # seconds
  poll_interval = 0.2  # seconds

  def __init__(
    self,
    port: str,
  ):
    if not HAS_SERIAL:
      raise RuntimeError(
        "pyserial is not installed. Install with: pip install pylabrobot[serial]. "
        f"Import error: {_SERIAL_IMPORT_ERROR}"
      )
    super().__init__()

    # BL-1300 Barcode reader factory default serial communication settings
    # should be the same factory default for the BL-600HA and BL-1300 models
    self.io = Serial(
      human_readable_device_name="Keyence Barcode Scanner",
      port=port,
      baudrate=self.default_baudrate,
      bytesize=serial.SEVENBITS,
      parity=serial.PARITY_EVEN,
      stopbits=serial.STOPBITS_ONE,
      write_timeout=1,
      timeout=1,
      rtscts=False,
    )

  async def setup(self):
    await self.io.setup()
    await self.initialize()

  async def initialize(self):
    """Initialize the Keyence barcode scanner."""

    deadline = time.time() + self.init_timeout
    while time.time() < deadline:
      response = await self.send_command("RMOTOR")
      if response.strip() == "MOTORON":
        logger.info("Barcode scanner motor is ON.")
        break
      elif response.strip() == "MOTOROFF":
        raise BarcodeScannerError("Failed to initialize Keyence barcode scanner: Motor is off.")
      await asyncio.sleep(self.poll_interval)
    else:
      raise BarcodeScannerError(
        "Failed to initialize Keyence barcode scanner: Timeout waiting for motor to turn on."
      )

  async def send_command(self, command: str) -> str:
    """Send a command to the barcode scanner and return the response.
    Keyence uses carriage return \r as the line ending by default."""

    await self.io.write((command + "\r").encode(self.serial_messaging_encoding))
    response = await self.io.read()
    return response.decode(self.serial_messaging_encoding).strip()

  async def stop(self):
    await self.io.stop()

  async def scan_barcode(self) -> Barcode:
    data = await self.send_command("LON")
    if data.startswith("NG"):
      raise BarcodeScannerError("Barcode reader is off: cannot read barcode")
    if data.startswith("ERR99"):
      raise BarcodeScannerError(f"Error response from barcode reader: {data}")
    return Barcode(data=data, symbology="unknown", position_on_resource="front")



import textwrap
from typing import List, Optional

from pylabrobot.resources.coordinate import Coordinate
from pylabrobot.resources.deck import Deck
from pylabrobot.resources.resource import Resource
from pylabrobot.resources.trash import Trash


class OTDeck(Deck):
  """The OpenTron deck."""

  def __init__(
    self,
    size_x: float = 624.3,
    size_y: float = 565.2,
    size_z: float = 900,
    origin: Coordinate = Coordinate(0, 0, 0),
    with_trash: bool = True,
    name: str = "deck",
  ):
    # size_z is probably wrong

    super().__init__(size_x=size_x, size_y=size_y, size_z=size_z, origin=origin)

    self.slots: List[Optional[Resource]] = [None] * 12

    self.slot_locations = [
      Coordinate(x=0.0, y=0.0, z=0.0),
      Coordinate(x=132.5, y=0.0, z=0.0),
      Coordinate(x=265.0, y=0.0, z=0.0),
      Coordinate(x=0.0, y=90.5, z=0.0),
      Coordinate(x=132.5, y=90.5, z=0.0),
      Coordinate(x=265.0, y=90.5, z=0.0),
      Coordinate(x=0.0, y=181.0, z=0.0),
      Coordinate(x=132.5, y=181.0, z=0.0),
      Coordinate(x=265.0, y=181.0, z=0.0),
      Coordinate(x=0.0, y=271.5, z=0.0),
      Coordinate(x=132.5, y=271.5, z=0.0),
      Coordinate(x=265.0, y=271.5, z=0.0),
    ]

    if with_trash:
      self._assign_trash()

  def _assign_trash(self):
    """Assign the trash area to the deck.

    Because all opentrons operations require that the resource passed references a parent, we need
    to create a dummy resource to represent the container of the actual trash area.
    """

    trash_container = Trash(
      name="trash_container",
      size_x=172.86,
      size_y=165.86,
      size_z=82,
    )

    actual_trash = Trash(
      name="trash",
      size_x=172.86,
      size_y=165.86,
      size_z=82,
    )

    # Trash location used to be Coordinate(x=86.43, y=82.93, z=0),
    # this is approximately the center of the trash area.
    # LiquidHandler will now automatically find the center of the trash before discarding tips,
    # so this location is no longer needed and we just use Coordinate.zero().
    # The actual location of the trash is determined by the slot number (12).
    trash_container.assign_child_resource(actual_trash, location=Coordinate.zero())
    self.assign_child_at_slot(trash_container, 12)

  def assign_child_resource(
    self,
    resource: Resource,
    location: Optional[Coordinate],
    reassign: bool = True,
  ):
    """Assign a resource to a slot.

    ..warning:: This method exists only for deserialization. You should use
    :meth:`assign_child_at_slot` instead.
    """

    if location is None:
      raise ValueError("location must be provided for resources on the deck")

    if location not in self.slot_locations:
      super().assign_child_resource(resource, location=location)
    else:
      slot = self.slot_locations.index(location) + 1
      self.assign_child_at_slot(resource, slot)

  def assign_child_at_slot(self, resource: Resource, slot: int):
    if slot not in range(1, 13):
      raise ValueError("slot must be between 1 and 12")

    if self.slots[slot - 1] is not None:
      raise ValueError(f"Spot {slot} is already occupied")

    self.slots[slot - 1] = resource
    super().assign_child_resource(resource, location=self.slot_locations[slot - 1])

  def unassign_child_resource(self, resource: Resource):
    if resource not in self.slots:
      raise ValueError(f"Resource {resource.name} is not assigned to this deck")

    slot = self.slots.index(resource)
    self.slots[slot] = None
    super().unassign_child_resource(resource)

  def get_slot(self, resource: Resource) -> Optional[int]:
    """Get the slot number of a resource."""
    if resource not in self.slots:
      return None
    return self.slots.index(resource) + 1

  def summary(self) -> str:
    """Get a summary of the deck.

    >>> print(deck.summary())

    Deck: 624.3mm x 565.2mm

    +-----------------+-----------------+-----------------+
    |                 |                 |                 |
    | 10: Empty       | 11: Empty       | 12: Trash       |
    |                 |                 |                 |
    +-----------------+-----------------+-----------------+
    |                 |                 |                 |
    |  7: tip_rack_1  |  8: tip_rack_2  |  9: tip_rack_3  |
    |                 |                 |                 |
    +-----------------+-----------------+-----------------+
    |                 |                 |                 |
    |  4: my_plate    |  5: my_other... |  6: Empty       |
    |                 |                 |                 |
    +-----------------+-----------------+-----------------+
    |                 |                 |                 |
    |  1: Empty       |  2: Empty       |  3: Empty       |
    |                 |                 |                 |
    +-----------------+-----------------+-----------------+
    """

    def _get_slot_name(slot: int) -> str:
      """Get slot name, or 'Empty' if slot is empty. If the name is too long, truncate it."""
      length = 11
      resource = self.slots[slot]
      if resource is None:
        return "Empty".ljust(length)
      name = resource.name
      if len(name) > 10:
        name = name[:8] + "..."
      return name.ljust(length)

    summary_ = f"""
      Deck: {self.get_absolute_size_x()}mm x {self.get_absolute_size_y()}mm

      +-----------------+-----------------+-----------------+
      |                 |                 |                 |
      | 10: {_get_slot_name(9)} | 11: {_get_slot_name(10)} | 12: {_get_slot_name(11)} |
      |                 |                 |                 |
      +-----------------+-----------------+-----------------+
      |                 |                 |                 |
      |  7: {_get_slot_name(6)} |  8: {_get_slot_name(7)} |  9: {_get_slot_name(8)} |
      |                 |                 |                 |
      +-----------------+-----------------+-----------------+
      |                 |                 |                 |
      |  4: {_get_slot_name(3)} |  5: {_get_slot_name(4)} |  6: {_get_slot_name(5)} |
      |                 |                 |                 |
      +-----------------+-----------------+-----------------+
      |                 |                 |                 |
      |  1: {_get_slot_name(0)} |  2: {_get_slot_name(1)} |  3: {_get_slot_name(2)} |
      |                 |                 |                 |
      +-----------------+-----------------+-----------------+
    """

    return textwrap.dedent(summary_)



from __future__ import annotations

import warnings
from dataclasses import dataclass
from typing import Callable, Optional

from pylabrobot.resources.volume_tracker import VolumeTracker
from pylabrobot.serializer import SerializableMixin


@dataclass
class Tip(SerializableMixin):
  """A single tip.

  Attributes:
    has_filter: whether the tip type has a filter
    total_tip_length: total length of the tip, in in mm
    maximal_volume: maximal volume of the tip, in ul
    fitting_depth: the overlap between the tip and the pipette, in mm
    name: optional identifier for this tip
  """

  has_filter: bool
  total_tip_length: float
  maximal_volume: float
  fitting_depth: float
  name: Optional[str] = None

  def __post_init__(self):
    if self.name is None:
      warnings.warn(
        "Creating a Tip without a name is deprecated. "
        "Tips created from deck resources (e.g. TipSpot) should be named.",
        DeprecationWarning,
        stacklevel=2,
      )

    thing = self.name or "tip_tracker"
    self.tracker = VolumeTracker(thing=thing, max_volume=self.maximal_volume)

  def serialize(self) -> dict:
    return {
      "type": self.__class__.__name__,
      "name": self.name,
      "total_tip_length": self.total_tip_length,
      "has_filter": self.has_filter,
      "maximal_volume": self.maximal_volume,
      "fitting_depth": self.fitting_depth,
    }

  def __hash__(self):
    return hash(
      (
        self.has_filter,
        self.total_tip_length,
        self.maximal_volume,
        self.fitting_depth,
      )
    )

  def __eq__(self, other: object) -> bool:
    if not isinstance(other, Tip):
      return NotImplemented

    return (
      self.has_filter == other.has_filter
      and self.total_tip_length == other.total_tip_length
      and self.maximal_volume == other.maximal_volume
      and self.fitting_depth == other.fitting_depth
    )


TipCreator = Callable[[str], Tip]
